"""Module for fitting timestamp differences with Gaussian function.

This file can also be imported as a module and contains the following
functions:

    * fit_with_gaussian - fit timestamp differences of a pair of pixels
    with a Gaussian function and plot a histogram of timestamp
    differences and the fit in a single figure.

    * fit_with_gaussian_all - find all peaks above the given threshold 
    and fit them individually with a Gaussian function.

    * fit_with_gaussian_full_sensor - fit timestamp differences of a
    pair of pixels (one from each half of the sensor) with a Gaussian
    function and plot a histogram of timestamp differences and the fit
    in a single figure.

    * fit_with_gaussian_fancy - fit timestamp diferences of a pair of
    pixels using the lmfit library. The main parameters reported are 
    the standard deviation, mean value, and contrast, together with
    residuals and signal-to-noise ratio (SNR) defined as a ratio of peak
    height to standard deviation of background.
    
    * unpickle_fit - unpickle the '.pkl' file, show the plot, and
    return the plot data for each line found (histogram, fit 1, fit 2, 
    etc.).

"""

import glob
import os
import pickle
import re
from typing import List

import numpy as np
import pandas as pd
from lmfit.models import GaussianModel, LinearModel
from matplotlib import pyplot as plt
from pandas import DataFrame
from pyarrow import feather as ft
from scipy import signal as sg

from LinoSPAD2.functions import utils


# TODO
def fit_with_gaussian(
    path: str,
    pixels: List[int] | List[List[int]],
    ft_file: str = None,
    window: float = 5e3,
    multiplier: int = 1,
    color_data: str = "rebeccapurple",
    color_fit: str = "darkorange",
    title_on: bool = True,
    correct_pix_address: bool = False,
    return_fit_params: bool = False,
    pickle_figure: bool = False,
) -> DataFrame:
    """Fit with Gaussian function and plot it.

    Fits timestamp differences of a pair of pixels with Gaussian
    function and plots it next to the histogram of the differences.
    Timestamp differences are collected from a '.feather' file with
    those if such exists.

    Parameters
    ----------
    path : str
        Path to the folder with '.dat' data files or where the
        '.feather' file with timestamp differences is located.
    pixels : List[int] | List[List[int]]
        List of pixel numbers for which the timestamp differences should
        be calculated and saved or list of two lists with pixel numbers
        for peak vs. peak calculations.
    ft_file : str, optional
        Name of the '.feather' file to use for plotting. Can be used
        when the raw '.dat' data is not available. The default is None.
    window : float, optional
        Time range in which timestamp differences are fitted. The
        default is 5e3.
    multiplier : int, optional
        Bins of delta t histogram should be in units of 17.857 (average
        LinoSPAD2 TDC bin width), this parameter helps with changing the
        bin size while maintaining that rule. Default is 1.
    color_data : str, optional
        For changing the color of the data. The default is "rebeccapurple".
    color_fit : str, optional
        For changing the color of the fit. The default is "darkorange".
    title_on : bool, optional
        Switch for turning on/off the title of the plot, the title
        shows the pixels for which the fit is done. The default is True.
    correct_pix_address : bool, optional
        Correct pixel address for the FPGA board on side 23. The
        default is False.
    return_fit_params : bool, optional
        Switch for returning the fit parameters for further analysis.
        The default is False.
    pickle_figure : bool, optional
        Switch for pickling the plot. Can be used to extract the plot
        data. The default is False.

    Raises
    ------
    FileNotFoundError
        Raised when no '.dat' data files are found.
    FileNotFoundError
        Raised when no '.feather' file with timestamp differences is found.
    ValueError
        Raised when no data for the requested pair of pixels was found
        in the '.feather' file.

    Returns
    -------
    DataFrame.
        A dataframe with fit parameters and their standard errors.
        Returned only if the "return_fit_params" is set to True.

    """
    plt.ion()

    os.chdir(path)

    # Correct pixel addressing for motherboard on side '23'
    if correct_pix_address:
        pixels = utils.correct_pixels_address(pixels)

    # Handle the input pixel list
    pixels_left, pixels_right = utils.pixel_list_transform(pixels)

    # If name of the '.feather' file is not provided, get it based on
    # the '.dat' files in the folder
    if ft_file is not None:
        file_name = ft_file.split(".")[0]
        feather_file_name = ft_file
    else:
        files = sorted(glob.glob("*.dat*"))
        file_name = files[0][:-4] + "-" + files[-1][:-4]

        try:
            os.chdir("delta_ts_data")
        except FileNotFoundError:
            raise ("\nFile with data not found")

        feather_file_name = glob.glob(f"*{file_name}.feather*")[0]

    if not feather_file_name:
        raise FileNotFoundError("\nFile with data not found")

    if return_fit_params:
        fit_params = {}

    data = ft.read_feather(f"{feather_file_name}")

    for pix_left in pixels_left:
        for pix_right in pixels_right:
            data_to_plot = data[f"{pix_left},{pix_right}"].dropna()

            # Check if there any finite values
            if not np.any(~np.isnan(data_to_plot)):
                raise ValueError(
                    "\nNo data for the requested pixel pair available"
                )

            data_to_plot = data_to_plot.dropna()
            data_to_plot = np.array(data_to_plot)

            # Use the given window for trimming the data for fitting
            data_to_plot = np.delete(
                data_to_plot, np.argwhere(data_to_plot < -window)
            )
            data_to_plot = np.delete(
                data_to_plot, np.argwhere(data_to_plot > window)
            )

            os.chdir(path)
            plt.rcParams.update({"font.size": 27})

            # Bins must be in units of 17.857 ps (2500/140)
            bins = np.arange(
                np.min(data_to_plot),
                np.max(data_to_plot),
                2500 / 140 * multiplier * 2,
            )

            # Calculate histogram of timestamp differences for primary guess
            # of fit parameters and selecting a narrower window for the fit
            n, b = np.histogram(data_to_plot, bins)

            try:
                n_argmax = np.argmax(n)
            except ValueError:
                print("Couldn't find position of histogram max")

            data_to_plot = np.delete(
                data_to_plot,
                np.argwhere(data_to_plot < b[n_argmax] - window / 2),
            )
            data_to_plot = np.delete(
                data_to_plot,
                np.argwhere(data_to_plot > b[n_argmax] + window / 2),
            )

            # Bins must be in units of 17.857 ps (2500/140)
            bins = np.arange(
                np.min(data_to_plot),
                np.max(data_to_plot),
                2500 / 140 * multiplier,
            )

            n, b = np.histogram(data_to_plot, bins)

            bin_centers = (b - 2500 / 140 * multiplier / 2)[1:]

            par, pcov = utils.fit_gaussian(bin_centers, n)

            # Interpolate for smoother fit plot
            to_fit_b = np.linspace(
                np.min(bin_centers),
                np.max(bin_centers),
                len(bin_centers) * 100,
            )
            to_fit_n = utils.gaussian(to_fit_b, par[0], par[1], par[2], par[3])

            perr = np.sqrt(np.diag(pcov))
            contrast = par[0] / par[3] * 100
            contrast_error = utils.error_propagation_division(
                par[0], perr[0], par[3], perr[3]
            )

            # Contrast error in %
            contrast_error = contrast_error * 100

            # Prepare dataframe for fit parameters, if return of them is
            # requested
            if return_fit_params:
                params_df = pd.DataFrame()
                params_df["Fit parameter"] = [
                    "center (ps)",
                    "center_error (ps)",
                    "sigma (ps)",
                    "sigma_error (ps)",
                    "contrast (%)",
                    "contrast_error (%)",
                ]
                params_df[f"Peak at {par[1]:.0f} ps"] = [
                    par[1],
                    perr[1],
                    par[2],
                    perr[2],
                    contrast,
                    contrast_error,
                ]
                fit_params[f"{pix_left},{pix_right}"] = params_df

            fig = plt.figure(figsize=(16, 10))
            plt.locator_params(axis="x", nbins=5)
            plt.xlabel(r"$\Delta$t (ps)")
            plt.ylabel("# of coincidences (-)")
            plt.step(
                b[1:],
                n,
                color=color_data,
                label="data",
            )
            plt.plot(
                to_fit_b,
                to_fit_n,
                "-",
                color=color_fit,
                label="fit\n"
                "\u03C3=({p1}\u00B1{pe1}) ps\n"
                "\u03BC=({p2}\u00B1{pe2}) ps\n"
                "C=({contrast}\u00B1{contrast_error}) %\n"
                "bkg={bkg}\u00B1{bkg_er}".format(
                    p1=format(par[2], ".0f"),
                    p2=format(par[1], ".0f"),
                    pe1=format(perr[2], ".0f"),
                    pe2=format(perr[1], ".0f"),
                    bkg=format(par[3], ".0f"),
                    bkg_er=format(perr[3], ".0f"),
                    contrast=format(contrast, ".1f"),
                    contrast_error=format(contrast_error, ".1f"),
                ),
            )
            plt.legend(loc="best")
            if title_on is True:
                plt.title(
                    "Gaussian fit of delta t histogram, pixels "
                    f"{pix_left}, {pix_right}"
                )

            try:
                os.chdir("results/fits")
            except FileNotFoundError:
                os.makedirs("results/fits")
                os.chdir("results/fits")

            # fig.tight_layout()  # for perfect spacing between the plots

            plt.savefig(f"{file_name}_pixels_{pix_left},{pix_right}_fit.png")

            # Pickle the figure if requested
            if pickle_figure:
                with open(
                    f"{file_name}_pixels_{pix_left},{pix_right}_fit.pkl", "wb"
                ) as f:
                    pickle.dump(fig, f)

            os.chdir("../..")

    return fit_params if return_fit_params else None


def fit_with_gaussian_all(
    path: str,
    pixels: List[int] | List[List[int]],
    ft_file: str = None,
    threshold_multiplier: float = 1.2,
    window: float = 5e3,
    multiplier: int = 1,
    color_d: str = "rebeccapurple",
    color_f: str = "darkorange",
    title_on: bool = True,
    correct_pix_address: bool = False,
    return_fit_params: bool = False,
    pickle_figure: bool = False,
) -> DataFrame:
    """Find all peaks above threshold and fit each with Gaussian.

    Finds all peaks above the given threshold (uses the 'threshold_multiplier'
    parameter as a multiplier of the median) and fits each with a
    Gaussian function.

    Parameters
    ----------
    path : str
        Path to the folder with '.dat' data files or where the
        '.feather' file with timestamp differences is located.
    pixels : List[int] | List[List[int]]
        List of pixel numbers for which the timestamp differences should
        be calculated and saved or list of two lists with pixel numbers
        for peak vs. peak calculations.
    ft_file : str, optional
        Name of the '.feather' file to use for plotting. Can be used
        when the raw '.dat' data is not available. The default is None.
    threshold_multiplier : float, optional
        Multiplier for the median of the histogram counts. Used for
        setting a threshold for the peak discovery in the histogram.
        The default is 1.2, or 20% above the median.
    window : float, optional
        Time range in which timestamp differences are fitted. The
        default is 5e3.
    multiplier : int, optional
        Bins of delta t histogram should be in units of 17.857 (average
        LinoSPAD2 TDC bin width). Default is 1.
    color_d : str, optional
        Color for the histogram. The default is "rebeccapurple".
    color_f : str, optional
        Color for the fit with Gaussian. The default is "darkorange".
    title_on : bool, optional
        Switch for showing the plot title. The default is "True".
    correct_pix_address : bool, optional
        Correct pixel address for the FPGA board on side 23. The
        default is False.
    return_fit_params : bool, optional
        Switch for returning the fit parameters for further analysis.
        The default is False.
    pickle_figure : bool, optional
        Switch for pickling the plot. Can be used to extract the plot
        data. The default is False.
    Raises
    ------
    FileNotFoundError
        Raised when no '.dat' data files are found.
    FileNotFoundError
        Raised when no '.feather' file with timestamp differences is
        found.
    ValueError
        Raised when no data for the requested pair of pixels was found
        in the '.feather' file.

    Returns
    -------
    DataFrame.
        A dataframe with fit parameters and their standard errors.
        Returned only if the "return_fit_params" is set to True.

    """

    os.chdir(path)

    # Correct pixel addressing for motherboard on side '23'
    if correct_pix_address:
        pixels = utils.correct_pixels_address(pixels)

    # Handle the input pixel list
    pixels_left, pixels_right = utils.pixel_list_transform(pixels)

    # If name of the '.feather' file is not provided, get it based on
    # the '.dat' files in the folder
    if ft_file is not None:
        file_name = ft_file.split(".")[0]
        feather_file_name = ft_file
    else:
        files = sorted(glob.glob("*.dat*"))
        file_name = files[0][:-4] + "-" + files[-1][:-4]

        try:
            os.chdir("delta_ts_data")
        except FileNotFoundError:
            raise ("\nFile with data not found")

        feather_file_name = glob.glob(f"*{file_name}.feather*")[0]

    if not feather_file_name:
        raise FileNotFoundError("\nFile with data not found")

    if return_fit_params:
        fit_params = {}

    data = ft.read_feather(
        f"{feather_file_name}",
    )

    for pix_left in pixels_left:
        for pix_right in pixels_right:

            data_to_plot = data[f"{pix_left},{pix_right}"].dropna()

            # Check if there any finite values
            if not np.any(~np.isnan(data_to_plot)):
                raise ValueError(
                    "\nNo data for the requested pixel pair available"
                )

            data_to_plot = data_to_plot.dropna()
            data_to_plot = np.array(data_to_plot)

            # Use the given window for trimming the data for fitting
            data_to_plot = np.delete(
                data_to_plot, np.argwhere(data_to_plot < -window)
            )
            data_to_plot = np.delete(
                data_to_plot, np.argwhere(data_to_plot > window)
            )

            os.chdir(path)
            plt.rcParams.update({"font.size": 27})

            # Bins must be in units of 17.857 ps (2500/140)
            bins = np.arange(
                np.min(data_to_plot),
                np.max(data_to_plot),
                2500 / 140 * multiplier,
            )

            # Calculate histogram of timestamp differences for primary guess
            # of fit parameters and selecting a narrower window for the fit
            n, b = np.histogram(data_to_plot, bins)

            bin_c = (b - (b[1] - b[0]) / 2)[1:]

            peak_pos = sg.find_peaks(
                n, height=np.median(n) * threshold_multiplier
            )[0]

            fig = plt.figure(figsize=(16, 10))
            plt.xlabel(r"$\Delta$t (ps)")
            plt.ylabel("# of coincidences (-)")
            plt.step(
                bin_c,
                n,
                color=color_d,
                label="data",
            )

            color_f = [
                "darkorange",
                "#ffa64e",
                "#c16b0f",
                "#ffbf81",
                "#874d13",
                "#ff9931",
                "#e07c09",
            ]
            plt.xticks(
                [-20e3, -10e3, 0, 10e3, 20e3],
                [
                    f"{-20e3:.0f}",
                    f"{-10e3:.0f}",
                    f"{0:.0f}",
                    f"{10e3:.0f}",
                    f"{20e3:.0f}",
                ],
            )
            labels = [
                "Peak 1",
                "Peak 2",
                "Peak 3",
                "Peak 4",
                "Peak 5",
                "Peak 6",
                "Peak 7",
            ]

            # Prepare dataframe for fit parameters, if return of them is
            # requested
            if return_fit_params:
                params_df = pd.DataFrame()
                for _ in peak_pos:
                    params_df["Fit parameter"] = [
                        "center (ps)",
                        "center_error (ps)",
                        "sigma (ps)",
                        "sigma_error (ps)",
                        "contrast (%)",
                        "contrast_error (%)",
                    ]

            for k, peak_ind in enumerate(peak_pos):
                data_to_fit = np.delete(
                    data_to_plot,
                    np.argwhere(data_to_plot < bin_c[peak_ind] - 3e3),
                )
                data_to_fit = np.delete(
                    data_to_fit,
                    np.argwhere(data_to_fit > bin_c[peak_ind] + 3e3),
                )

                # Bins must be in units of 17.857 ps (2500/140)
                bins = np.arange(
                    np.min(data_to_fit),
                    np.max(data_to_fit),
                    2500 / 140 * multiplier,
                )

                counts, bin_edges = np.histogram(data_to_fit, bins)

                bin_centers = (bin_edges - (bin_edges[1] - bin_edges[0]) / 2)[
                    1:
                ]

                par, pcov = utils.fit_gaussian(bin_centers, counts)

                # Interpolate for smoother fit plot
                to_fit_n1 = utils.gaussian(
                    bin_centers, par[0], par[1], par[2], par[3]
                )

                perr = np.sqrt(np.diag(pcov))

                contrast = par[0] / par[-1] * 100

                contrast_error = utils.error_propagation_division(
                    par[0], perr[0], par[3], perr[3]
                )
                contrast_error = contrast_error * 100

                # Add the fit parameters to the dataframe for returning
                # if requested
                if return_fit_params:
                    params_df[f"Peak at {par[1]:.0f} ps"] = [
                        par[1],
                        perr[1],
                        par[2],
                        perr[2],
                        contrast,
                        contrast_error,
                    ]

                plt.plot(
                    bin_centers,
                    to_fit_n1,
                    "-",
                    color=color_f[k],
                    linewidth=2,
                    label=f"{labels[k]}\n"
                    "\u03C3=({p1}\u00B1{pe1}) ps\n"
                    "\u03BC=({p2}\u00B1{pe2}) ps\n"
                    "C=({contrast}\u00B1{contrast_error}) %".format(
                        labels,
                        p1=format(par[2], ".0f"),
                        p2=format(par[1], ".0f"),
                        pe1=format(perr[2], ".0f"),
                        pe2=format(perr[1], ".0f"),
                        contrast=format(contrast, ".1f"),
                        contrast_error=format(contrast_error, ".1f"),
                    ),
                )
                lower_limit = par[1] - 2 * par[2]
                upper_limit = par[1] + 2 * par[2]

                data_in_interval = data_to_fit[
                    (data_to_fit >= lower_limit) & (data_to_fit <= upper_limit)
                ]

                bckg_center_position = par[1] - 7 * par[2]

                bckg_in_2sigma = data_to_fit[
                    (data_to_fit > bckg_center_position - 2 * par[2])
                    & (data_to_fit < bckg_center_position + 2 * par[2])
                ]

                # Plot the Gaussian fit and the 2-sigma interval
                er1 = np.sqrt(len(data_in_interval))
                er2 = np.sqrt(len(bckg_in_2sigma))

            if return_fit_params:
                fit_params[f"{pix_left},{pix_right}"] = params_df

            plt.xlim(-window, window)
            plt.legend(loc="best")
            if title_on is True:
                plt.title(
                    "Gaussian fit of delta t histogram, pixels "
                    f"{pix_left}, {pix_right}"
                )

            try:
                os.chdir("results/fits")
            except FileNotFoundError as _:
                os.makedirs("results/fits")
                os.chdir("results/fits")

            plt.savefig(
                f"{file_name}_pixels_{pix_left},{pix_right}_all_fit.png"
            )

            # Pickle the figure if requested
            if pickle_figure:
                with open(
                    f"{file_name}_pixels_{pix_left},{pix_right}_all_fit.pkl",
                    "wb",
                ) as f:
                    pickle.dump(fig, f)

            os.chdir("../..")

    return fit_params if return_fit_params else None


# TODO no data to test full sensor plot
# TODO add pickling plot
def fit_with_gaussian_full_sensor(
    path,
    pix_pair: List[int],
    window: float = 5e3,
    multiplier: int = 1,
    color_data: str = "rebeccapurple",
    color_fit: str = "darkorange",
    title_on: bool = True,
):
    """Fit with Gaussian function and plot it.

    Fits timestamp differences of a pair of pixels with Gaussian
    function and plots it next to the histogram of the differences.
    Timestamp differences are collected from a '.feather' file with
    those if such exists.

    Parameters
    ----------
    path : str
        Path to the folder with '.dat' data files.
    pix_pair : list
        Two pixel numbers for which fit is done.
    window : float, optional
        Time range in which timestamp differences are fitted. The
        default is 5e3.
    multiplier : int, optional
        Bins of delta t histogram should be in units of 17.857 (average
        LinoSPAD2 TDC bin width), this parameter helps with changing the
        bin size while maintaining that rule. Default is 1.
    color_data : str, optional
        For changing the color of the data. The default is "rebeccapurple".
    color_fit : str, optional
        For changing the color of the fit. The default is "darkorange".
    title_on : bool, optional
        Switch for turning on/off the title of the plot, the title
        shows the pixels for which the fit is done. The default is True.

    Raises
    ------
    FileNotFoundError
        Raised when no '.dat' data files are found.
    FileNotFoundError
        Raised when no '.feather' file with timestamp differences is found.
    ValueError
        Raised when no data for the requested pair of pixels was found
        in the '.feather' file.

    Returns
    -------
    None.

    """
    os.chdir(path)

    # Assuming the current directory contains the folders
    folders = glob.glob("*#*")

    # Ensure at least two folders are found
    if len(folders) < 2:
        raise ValueError(
            "At least two folders with the specified pattern are required."
        )

    # Depending on the order of the folders, the ".feather" file naming
    # could be different. The following code collects the names for
    # the two possible naming results

    # Collect ".dat" files' names in the first folder
    os.chdir(folders[0])
    files_all = sorted(glob.glob("*.dat*"))
    feather_file_name1 = files_all[0][:-4] + "-"
    feather_file_name2 = "-" + files_all[-1][:-4]

    # Collect ".dat" files' names in the second folder
    os.chdir(f"../{folders[1]}")
    files_all = sorted(glob.glob("*.dat"))
    feather_file_name1 += files_all[-1][:-4]
    feather_file_name2 = files_all[0][:-4] + feather_file_name2

    os.chdir("..")

    feather_file_path1 = f"delta_ts_data/{feather_file_name1}.feather"
    feather_file_path2 = f"delta_ts_data/{feather_file_name2}.feather"
    feather_file, feather_file_name = (
        (feather_file_path1, feather_file_name1)
        if os.path.isfile(feather_file_path1)
        else (feather_file_path2, feather_file_name2)
    )

    if not os.path.isfile(feather_file):
        raise FileNotFoundError(
            "'.feather' file with timestamps differences was not found"
        )

    data_to_plot = ft.read_feather(
        f"{feather_file}",
        columns=[f"{pix_pair[0]},{pix_pair[1]}"],
    ).dropna()

    # Check if there any finite values
    if not np.any(~np.isnan(data_to_plot)):
        raise ValueError("\nNo data for the requested pixel pair available")

    data_to_plot = data_to_plot.dropna()
    data_to_plot = np.array(data_to_plot)

    # Use the given window for trimming the data for fitting
    data_to_plot = np.delete(data_to_plot, np.argwhere(data_to_plot < -window))
    data_to_plot = np.delete(data_to_plot, np.argwhere(data_to_plot > window))

    os.chdir("..")
    plt.rcParams.update({"font.size": 27})

    # Bins must be in units of 17.857 ps (2500/140)
    bins = np.arange(
        np.min(data_to_plot), np.max(data_to_plot), 2500 / 140 * multiplier
    )

    # Calculate histogram of timestamp differences for primary guess
    # of fit parameters and selecting a narrower window for the fit
    n, b = np.histogram(data_to_plot, bins)

    try:
        n_argmax = np.argmax(n)
        cp_pos = (bins[n_argmax] + bins[n_argmax + 1]) / 2
    except ValueError:
        print("Couldn't find position of histogram max")

    data_to_plot = np.delete(
        data_to_plot, np.argwhere(data_to_plot < b[n_argmax] - window / 2)
    )
    data_to_plot = np.delete(
        data_to_plot, np.argwhere(data_to_plot > b[n_argmax] + window / 2)
    )

    # Bins must be in units of 17.857 ps (2500/140)
    bins = np.arange(
        np.min(data_to_plot), np.max(data_to_plot), 2500 / 140 * multiplier
    )

    n, b = np.histogram(data_to_plot, bins)

    bin_centers = (b - 2500 / 140 * multiplier / 2)[1:]

    par, pcov = utils.fit_gaussian(bin_centers, n)

    # Interpolate for smoother fit plot
    to_fit_b = np.linspace(
        np.min(bin_centers), np.max(bin_centers), len(bin_centers) * 100
    )
    to_fit_n = utils.gaussian(to_fit_b, par[0], par[1], par[2], par[3])

    perr = np.sqrt(np.diag(pcov))
    contrast = par[0] / par[3] * 100
    contrast_error = utils.error_propagation_division(
        par[0], perr[0], par[3], perr[3]
    )

    # Contrast error in %
    contrast_error = contrast_error * 100

    fig = plt.figure(figsize=(16, 10))
    plt.xlabel(r"$\Delta$t (ps)")
    plt.ylabel("# of coincidences (-)")
    plt.step(
        b[1:],
        n,
        color=color_data,
        label="data",
    )
    plt.plot(
        to_fit_b,
        to_fit_n,
        "-",
        color=color_fit,
        label="fit\n"
        "\u03C3=({p1}\u00B1{pe1}) ps\n"
        "\u03BC=({p2}\u00B1{pe2}) ps\n"
        "C=({contrast}\u00B1{vis_er}) %\n"
        "bkg={bkg}\u00B1{bkg_er}".format(
            p1=format(par[2], ".0f"),
            p2=format(par[1], ".0f"),
            pe1=format(perr[2], ".0f"),
            pe2=format(perr[1], ".0f"),
            bkg=format(par[3], ".0f"),
            bkg_er=format(perr[3], ".0f"),
            contrast=format(contrast, ".1f"),
            vis_er=format(contrast_error, ".1f"),
        ),
    )
    plt.legend(loc="best")
    if title_on is True:
        plt.title(
            "Gaussian fit of delta t histogram, pixels "
            f"{pix_pair[0]}, {pix_pair[1]}"
        )

    try:
        os.chdir(os.path.join(path, r"results/fits"))
    except FileNotFoundError as _:
        os.makedirs(os.path.join(path, r"results/fits"))
        os.chdir(os.path.join(path, r"results/fits"))

    fig.tight_layout()  # for perfect spacing between the plots

    plt.savefig(
        f"{feather_file_name}_pixels_{pix_pair[0]},{pix_pair[1]}_fit.png"
    )

    plt.pause(0.1)
    os.chdir("../..")


def fit_with_gaussian_fancy(
    path: str,
    pixels: List[int] | List[List[int]],
    ft_file: str = None,
    range_left: float = -5e3,
    range_right: float = 5e3,
    multiplier: int = None,
    return_fit_params: bool = False,
    interpolate_fit: bool = True,
    correct_pix_address: bool = False,
    pickle_figure: bool = False,
):
    """Fit with Gaussian function and plot it, using lmfit library.

    Fits timestamp differences of a pair of pixels with Gaussian
    function and plots it over to the histogram of the differences.
    Timestamp differences are collected from a '.feather' file.
    Additionally, the residuals are calculated and plotted below the
    data and fit, together with a normal distribution of them.

    Parameters
    ----------
    path : str
        Path to the folder with '.dat' data files or where the
        '.feather' file with timestamp differences is located.
    pixels : List[int] | List[List[int]]
        List of pixel numbers for which the timestamp differences should
        be calculated and saved or list of two lists with pixel numbers
        for peak vs. peak calculations.
    ft_file : str, optional
        Name of the '.feather' file with the timestamp differences.
        Should be provided when the data files are not available. The
        default is None.
    range_left : float, optional
        Left limit for the signal window. The default is -5e3.
    range_right : float, optional
        Right limit for the signal window. The default is 5e3.
    multiplier : int, optional
        Bins of delta t histogram should be in units of 17.857 (average
        LinoSPAD2 TDC bin width), this parameter helps with changing the
        bin size while maintaining that rule. Default is 1.
    return_fit_parameters : bool, optional
        Switch for returning the fit parameters as they are returned
        by the lmfit library. The default is False.
    interpolate_fit: bool, optional
        Switch for an interpolated fit. The default is True.
    correct_pix_address : bool, optional
        Correct pixel address for the FPGA board on side 23. The
        default is False.
    pickle_figure : bool, optional
        Switch for pickling the plot. Can be used to extract the plot
        data. The default is False.

    Returns
    -------
    lmfit.parameter.Parameters
        Parameters of the fit as returned by the lmfit library. Returned
        only if the "return_fit_params" is set to True.

    Raises
    ------
    FileNotFoundError
        Raised if the 'delta_ts_data' folder was not found.
    """

    os.chdir(path)

    # Correct pixel addressing for motherboard on side '23'
    if correct_pix_address:
        pixels = utils.correct_pixels_address(pixels)

    # Handle the input pixel list
    pixels_left, pixels_right = utils.pixel_list_transform(pixels)

    # If name of the '.feather' file is not provided, get it based on
    # the '.dat' files in the folder
    if ft_file is not None:
        file_name = ft_file.split(".")[0]
        feather_file_name = ft_file
    else:
        files = sorted(glob.glob("*.dat*"))
        file_name = files[0][:-4] + "-" + files[-1][:-4]

        try:
            os.chdir("delta_ts_data")
        except FileNotFoundError:
            raise FileNotFoundError(
                "Folder 'delta_ts_data' with the "
                "timestamps differences was not found."
            )

        feather_file_name = glob.glob(f"*{file_name}.feather*")[0]

    # # Select the column of delta ts for the requested pair of pixels
    # if pix_pair is not None:
    #     try:
    #         data = ft.read_feather(
    #             ft_file,
    #             columns=[f"{pix_pair[0]},{pix_pair[1]}"],
    #         )
    #     except ArrowInvalid:
    #         raise ArrowInvalid(
    #             "Data for the requested pair of pixels were not found"
    #         )
    # else:
    #     data = ft.read_feather(ft_file)
    #     pix_pair = [int(x) for x in data.columns[0].split(",")]

    if not feather_file_name:
        raise FileNotFoundError("\nFile with data not found")

    if return_fit_params:
        fit_params = {}

    data = ft.read_feather(f"{feather_file_name}")

    for pix_left in pixels_left:
        for pix_right in pixels_right:

            data_to_plot = data[f"{pix_left},{pix_right}"].dropna()

            # Select a window around the signal
            data_signal = data_to_plot[
                (data_to_plot >= range_left) & (data_to_plot <= range_right)
            ].dropna()

            # TODO - data problems - wider range for bckg
            # For SNR - can't use wider bckg range because of data problems
            # data_bckg = data[(data < range_left) | (data > range_right)].dropna()

            # n, _ = np.histogram(data_bckg, bins=200)
            # bckg_stderr = np.std(n)

            # Select background data for SNR calculation
            data_bckg = data_to_plot[
                (data_to_plot > range_right)
                & (data_to_plot < range_right + 10e3)
            ].dropna()

            n, _ = np.histogram(data_bckg, bins=200)

            # Normalize - same as signal
            n = n / np.median(n)

            # The divisor in the SNR (signal height / sigma of bckg)
            bckg_stderr = np.std(n)

            # Calculate histogram
            # Scott's rule for the number of bins

            if multiplier is None:
                number_of_bins = (
                    3.5
                    * np.std(data_signal.values)
                    / len(data_signal.values) ** (1 / 3)
                )
                multiplier = int(number_of_bins / (2500 / 140))
            bins = np.arange(
                np.min(data_signal),
                np.max(data_signal),
                2500 / 140 * multiplier,
            )
            counts, bins = np.histogram(data_signal, bins)

            # Normalize
            counts = counts / np.median(counts)

            bin_centers = (bins[:-1] + bins[1:]) / 2

            # Composite fit: Gaussian for the peak + linear for bckg of random
            # coincidences
            model_peak = GaussianModel()
            model_bckg = LinearModel()

            # Guess the initial values of parameters
            params_peak = model_peak.guess(counts, x=bin_centers)
            params_bckg = model_bckg.guess(counts, x=bin_centers)
            params = params_peak + params_bckg

            params["amplitude"].min = 0
            params["height"].min = 0
            params["height"].max = 1

            # Combine the models
            model = model_peak + model_bckg

            # Do the fitting
            result = model.fit(counts, params, x=bin_centers, max_nfev=1000)

            # For smoother fit
            if interpolate_fit:
                fit_bins_plot = np.linspace(
                    np.min(bin_centers),
                    np.max(bin_centers),
                    len(bin_centers) * 100,
                )
                fit_counts_plot = utils.gaussian(
                    fit_bins_plot,
                    result.params["height"].value,
                    result.params["center"].value,
                    result.params["sigma"].value,
                    result.params["intercept"].value,
                )

            # Plot results
            plt.rcParams.update({"font.size": 27})
            fig, ((ax1, _), (ax2, ax3)) = plt.subplots(
                2,
                2,
                figsize=(16, 10),
                gridspec_kw={"width_ratios": [3, 1], "height_ratios": [3, 1]},
            )

            # Data + fit
            ax1.plot(
                bin_centers, counts, ".", label="Data", color="rebeccapurple"
            )
            if not interpolate_fit:
                ax1.plot(
                    bin_centers,
                    result.best_fit,
                    label="Gaussian fit",
                    color="darkorange",
                )
            else:
                ax1.plot(
                    fit_bins_plot,
                    fit_counts_plot,
                    label="Gaussian fit",
                    color="darkorange",
                )
            ax1.set_ylabel("Counts (-)")
            # ax1.yaxis.set_label_coords(-0.105, 0.5)
            ax1.set_xticks([], [])
            ax1.legend()
            ax1.set_xlim(
                result.params["center"].value - 2.5e3,
                result.params["center"].value + 2.5e3,
            )

            # Parameters
            fit_params_text = "\n".join(
                [
                    "Fit parameters",
                    "                           ",
                    f"$\sigma$: ({result.params['sigma'].value:.0f}"
                    f"±{result.params['sigma'].stderr:.0f}) ps",
                    f"$\mu$: ({result.params['center'].value:.0f}"
                    f"±{result.params['center'].stderr:.0f}) ps",
                    f"C: ({result.params['height'].value*100:.0f}"
                    f"±{result.params['height'].stderr*100:.0f}) %",
                    f"SNR: {result.params['height'] / bckg_stderr:.0f} $\sigma$",
                ]
            )

            ax1.text(
                1.05,
                0.5,
                fit_params_text,
                transform=ax1.transAxes,
                fontsize=24,
                bbox=dict(
                    boxstyle="round,pad=0.5",
                    edgecolor="black",
                    facecolor="white",
                ),
            )

            # Plot the residuals
            result.plot_residuals(
                ax=ax2,
                title=" ",
                datafmt=".",
            )
            ax2.set_ylabel("Residuals (-)")
            ax2.set_xlabel("$\Delta$t (ps)")
            ax2_lines = ax2.get_lines()
            ax2_lines[0].set_color("black")
            ax2_lines[1].set_color("rebeccapurple")

            # Plot the distribution of residuals with a Gaussian fit
            residuals = counts - result.best_fit
            counts_residuals, bins_residuals = np.histogram(residuals, bins=20)
            bins_residuals_edges = (
                bins_residuals[:-1] + bins_residuals[1:]
            ) / 2
            ax3.plot(
                counts_residuals,
                bins_residuals_edges,
                ".",
                color="rebeccapurple",
            )

            model_residuals = GaussianModel()
            params_residuals = model_residuals.guess(
                counts_residuals, x=bins_residuals_edges
            )
            result_residuals = model_residuals.fit(
                counts_residuals, params_residuals, x=bins_residuals_edges
            )

            ax3.plot(
                result_residuals.best_fit,
                bins_residuals_edges,
                color="darkorange",
                label="\n".join(
                    [
                        f"$\sigma$: {result_residuals.params['sigma'].value:.2f}",
                        f"$\mu$: {round(result_residuals.params['center'].value, 3):.2f}",
                    ]
                ),
            )
            y_limits = ax2.get_ylim()
            ax3.set_ylim(y_limits)
            ax3.set_yticks([], [])
            ax3.legend(loc="best", fontsize=15)

            fig.delaxes(_)
            plt.tight_layout()
            plt.subplots_adjust(hspace=0.05)
            plt.subplots_adjust(wspace=0.05)

            # Plot the report of the data fit
            print(result.fit_report())

            # Save the plot
            try:
                os.chdir(os.path.join(path, r"results/fits"))
            except FileNotFoundError as _:
                os.makedirs(os.path.join(path, r"results/fits"))
                os.chdir(os.path.join(path, r"results/fits"))

            plot_name = ft_file.split(".")[0]
            plt.savefig(
                f"{plot_name}_pixels_{pix_left},{pix_right}_fancy_fit.png"
            )

            # Pickle the figure if requested
            if pickle_figure:
                with open(
                    f"{file_name}_pixels_{pix_left},{pix_right}_fancy_fit.pkl",
                    "wb",
                ) as f:
                    pickle.dump(fig, f)

            if return_fit_params:
                fit_params[f"{pix_left},{pix_right}"] = result.params

    os.chdir("../..")

    if return_fit_params:
        return fit_params


def unpickle_fit(fit_pickle_file: str) -> dict:
    """Unpickle a saved figure and return plot data.

    Load a pickled figure, extract and return the histogram and fit
    data.

    Parameters
    ----------
    fit_pickle_file : str
        The absolute path to the pickle file with the fit plot.

    Returns
    -------
    tuple
        A tuple containing:
        - fig (matplotlib.figure.Figure): The unpickled figure object.
        - plot_data (dict): A dictionary with keys "Histogram" and
        "Fit_{i}" (for i >= 1), where each value is a tuple of x and y
        data for the corresponding line plot.
        - params_df (pandas.DataFrame): A DataFrame containing fit
        parameters extracted from the legend. Columns represent each
        fit.

    Raises
    ------
    FileNotFoundError
        If the specified pickle file does not exist, a FileNotFoundError
        is raised and an error message is printed.
    """
    try:
        with open(fit_pickle_file, "rb") as f:
            fig = pickle.load(f)
    except FileNotFoundError as e:
        print(f" {e}")

    # Pack the data into a dictionary, first is the histogram, others
    # are the fits
    plot_data = {}

    # Extract the data; go over lines, stack them into dictionary
    ax = fig.axes[0]
    for i, line in enumerate(ax.lines):
        x, y = line.get_data()
        if i == 0:
            plot_data["Histogram"] = (x, y)
        else:
            plot_data[f"Fit_{i}"] = (x, y)

    # Extract the legend
    legend = ax.get_legend()

    legend_labels = [text.get_text() for text in legend.get_texts()]

    # Pattern for the fit parameters: sigma, mu, C
    pattern = (
        r"σ=\(([\d.-]+)±([\d.-]+)\) ps\n"
        r"μ=\(([\d.-]+)±([\d.-]+)\) ps\n"
        r"C=\(([\d.-]+)±([\d.-]+)\) %"
    )

    # Prepare a dataframe for the fit parameters
    params_df = pd.DataFrame()

    params_df["Fit parameter"] = [
        "center (ps)",
        "center_error (ps)",
        "sigma (ps)",
        "sigma_error (ps)",
        "contrast (%)",
        "contrast_error (%)",
    ]

    # Go over labels and collect the parameters
    for label in legend_labels:
        # Search for the fit parameters in the legend label
        match = re.search(pattern, label)

        if match:
            # Extract the fit parameters and their uncertainties
            sigma, sigma_err = match.group(1), match.group(2)
            mu, mu_err = match.group(3), match.group(4)
            C, C_err = match.group(5), match.group(6)

            # Fill the dataframe
            params_df[f"Peak at {mu} ps"] = [
                mu,
                mu_err,
                sigma,
                sigma_err,
                C,
                C_err,
            ]

    return fig, plot_data, params_df



"""Module for computing timestamp differences.

Compares all timestamps from the same cycle for the given pair of pixels
against a given value (delta_window). Differences in that window are saved
and returned as a list.

This file can also be imported as a module and contains the following
functions:

    TODO remove
    * calculate_differences_2212 - calculate timestamp differences for
    the given pair of pixels. Works only with firmware version '2212'.
    
    * calculate_differences_2212_fast - calculate timestamp differences for
    the given pair of pixels. Works only with firmware version '2212'.
    Uses a faster algorithm than the function above.

"""

from typing import List
from warnings import warn

import numpy as np
import pandas as pd
from numpy import ndarray

from LinoSPAD2.functions import utils


def calculate_differences_2212(
    data: List[float],
    pixels: List[int] | List[List[int]],
    pix_coor,
    delta_window: float = 50e3,
):
    """Calculate timestamp differences for firmware version 2212.

    Calculate timestamp differences for the given pixels and LinoSPAD2
    firmware version 2212.

    Parameters
    ----------
    data : list of array-like
        List of data arrays, each corresponding to a different TDC.
    pixels : List[int] | List[List[int]]
        List of pixel numbers for which the timestamp differences should
        be calculated or list of two lists with pixel numbers for peak
        vs. peak calculations.
    pix_coor : array-like
        Array for transforming the pixel address in terms of TDC (0 to 3)
        to pixel number in terms of half of the sensor (0 to 255).
    delta_window : float, optional
        Width of the time window for counting timestamp differences.
        The default is 50e3 (50 ns).

    Returns
    -------
    deltas_all : dict
        Dictionary containing timestamp differences for each pair of pixels.

    """

    # TODO: remove
    warn(
        "This function is deprecated. Use" "'calculate_differences_2212_fast'",
        DeprecationWarning,
        stacklevel=2,
    )
    # Dictionary for the timestamp differences, where keys are the
    # pixel numbers of the requested pairs
    deltas_all = {}

    pixels_left, pixels_right = utils.pixel_list_transform(pixels)

    # Find ends of cycles
    cycle_ends = np.argwhere(data[0].T[0] == -2)
    cycle_ends = np.insert(cycle_ends, 0, 0)

    for q in pixels_left:
        # First pixel in the pair
        tdc1, pix_c1 = np.argwhere(pix_coor == q)[0]
        pix1 = np.where(data[tdc1].T[0] == pix_c1)[0]
        for w in pixels_right:
            if w <= q:
                continue
            deltas_all[f"{q},{w}"] = []

            # Second pixel in the pair
            tdc2, pix_c2 = np.argwhere(pix_coor == w)[0]
            pix2 = np.where(data[tdc2].T[0] == pix_c2)[0]

            # Go over cycles, getting data for the appropriate cycle
            # only
            for cyc in range(len(cycle_ends) - 1):
                slice_from = cycle_ends[cyc]
                slice_to = cycle_ends[cyc + 1]
                pix1_slice = pix1[(pix1 >= slice_from) & (pix1 < slice_to)]
                if not np.any(pix1_slice):
                    continue
                pix2_slice = pix2[(pix2 >= slice_from) & (pix2 < slice_to)]
                if not np.any(pix2_slice):
                    continue

                # Calculate delta t
                tmsp1 = data[tdc1].T[1][pix1_slice]
                tmsp1 = tmsp1[tmsp1 > 0]
                tmsp2 = data[tdc2].T[1][pix2_slice]
                tmsp2 = tmsp2[tmsp2 > 0]
                for t1 in tmsp1:
                    deltas = tmsp2 - t1
                    ind = np.where(np.abs(deltas) < delta_window)[0]
                    deltas_all[f"{q},{w}"].extend(deltas[ind])

    return deltas_all


def calculate_differences_2212_fast(
    data: ndarray,
    pixels: List[int] | List[List[int]],
    pix_coor: ndarray,
    delta_window: float = 50e3,
    cycle_length: float = 4e9,
):
    """Calculate timestamp differences for firmware version 2212.

    Calculate timestamp differences for the given pixels and LinoSPAD2
    firmware version 2212.

    Parameters
    ----------
    data : ndarray
        Matrix of timestamps, where rows correspond to the TDCs.
    pixels : List[int] | List[List[int]]
        List of pixel numbers for which the timestamp differences should
        be calculated or list of two lists with pixel numbers for peak
        vs. peak calculations.
    pix_coor : ndarray
        Array for transforming the pixel address in terms of TDC (0 to 3)
        to pixel number in terms of half of the sensor (0 to 255).
    delta_window : float, optional
        Width of the time window for counting timestamp differences.
        The default is 50e3 (50 ns).
    cycle_length : float, optional
        Length of each acquisition cycle. The default is 4e9 (4 ms).

    Returns
    -------
    deltas_all : dict
        Dictionary containing timestamp differences for each pair of pixels.

    """

    # Dictionary for the timestamp differences, where keys are the
    # pixel numbers of the requested pairs
    deltas_all = {}

    pixels_left, pixels_right = utils.pixel_list_transform(pixels)

    # Find ends of cycles
    cycle_ends = np.argwhere(data[0].T[0] == -2)
    cycle_ends = np.insert(cycle_ends, 0, 0)

    for q in pixels_left:
        # First pixel in the pair
        tdc1, pix_c1 = np.argwhere(pix_coor == q)[0]
        pix1 = np.where(data[tdc1].T[0] == pix_c1)[0]
        for w in pixels_right:
            if w <= q:
                continue
            deltas_all[f"{q},{w}"] = []

            timestamps_1 = []
            timestamps_2 = []

            # Second pixel in the pair
            tdc2, pix_c2 = np.argwhere(pix_coor == w)[0]
            pix2 = np.where(data[tdc2].T[0] == pix_c2)[0]

            # Go over cycles, shifting the timestamps from each next
            # cycle by lengths of cycles before (e.g., for the 4th cycle
            # add 12 ms)
            for i, _ in enumerate(cycle_ends[:-1]):
                slice_from = cycle_ends[i]
                slice_to = cycle_ends[i + 1]
                pix1_slice = pix1[(pix1 >= slice_from) & (pix1 < slice_to)]
                if not np.any(pix1_slice):
                    continue
                pix2_slice = pix2[(pix2 >= slice_from) & (pix2 < slice_to)]
                if not np.any(pix2_slice):
                    continue

                # Shift timestamps by cycle length
                tmsp1 = data[tdc1].T[1][pix1_slice]
                tmsp1 = tmsp1[tmsp1 > 0]
                tmsp1 = tmsp1 + cycle_length * i

                tmsp2 = data[tdc2].T[1][pix2_slice]
                tmsp2 = tmsp2[tmsp2 > 0]
                tmsp2 = tmsp2 + cycle_length * i

                timestamps_1.extend(tmsp1)
                timestamps_2.extend(tmsp2)

            timestamps_1 = np.array(timestamps_1)
            timestamps_2 = np.array(timestamps_2)

            # Indicators for each pixel: 0 for timestamps from one pixel
            # 1 - from the other
            pix1_ind = np.zeros(len(timestamps_1), dtype=np.int32)
            pix2_ind = np.ones(len(timestamps_2), dtype=np.int32)

            pix1_data = np.vstack((pix1_ind, timestamps_1))
            pix2_data = np.vstack((pix2_ind, timestamps_2))

            # Dataframe for each pixel with pixel indicator and
            # timestamps
            df1 = pd.DataFrame(
                pix1_data.T, columns=["Pixel_index", "Timestamp"]
            )
            df2 = pd.DataFrame(
                pix2_data.T, columns=["Pixel_index", "Timestamp"]
            )

            # Combine the two dataframes
            df_combined = pd.concat((df1, df2), ignore_index=True)

            # Sort the timestamps
            df_combined.sort_values("Timestamp", inplace=True)

            # Subtract pixel indicators of neighbors; values of 0
            # correspond to timestamp differences for the same pixel
            # '-1' and '1' - to differences from different pixels
            df_combined["Pixel_index_diff"] = df_combined["Pixel_index"].diff()

            # Calculate timestamp difference between neighbors
            df_combined["Timestamp_diff"] = df_combined["Timestamp"].diff()

            # Get the correct timestamp difference sign
            df_combined["Timestamp_diff"] = (
                df_combined["Timestamp_diff"] * df_combined["Pixel_index_diff"]
            )

            # Collect timestamp differences where timestamps are from
            # different pixels
            filtered_df = df_combined[
                abs(df_combined["Pixel_index_diff"]) == 1
            ]

            # Save only timestamps differences in the requested window
            delta_ts = filtered_df[
                abs(filtered_df["Timestamp_diff"]) < delta_window
            ]["Timestamp_diff"].values

            deltas_all[f"{q},{w}"].extend(delta_ts)

    return deltas_all



"""
Module with scripts for plotting the LinoSPAD2 sensor population.

This script utilizes an unpacking module used specifically for the
LinoSPAD2 data output.

This file can also be imported as a module and contains the following
functions:

    * collect_data_and_apply_mask - Collect data from files and apply
    mask to the valid pixel count.

    * plot_single_pix_hist - Plot a histogram for each pixel in the given
    range.

    * plot_sensor_population - Plot number of timestamps in each pixel
    for all data files.

    * plot_sensor_population_spdc - Plot sensor population for SPDC data.

    * plot_sensor_population_full_sensor - Plot the number of timestamps
    in each pixel for all data files from two different FPGAs/sensor
    halves.
"""

import glob
import os
import pickle
import sys
from typing import List

import numpy as np
from matplotlib import pyplot as plt
from scipy.optimize import curve_fit
from scipy.signal import find_peaks
from tqdm import tqdm

from LinoSPAD2.functions import unpack as f_up
from LinoSPAD2.functions import utils


# TODO inc offset and apply calibration - remove
def collect_data_and_apply_mask(
    files: List[str],
    daughterboard_number: str,
    motherboard_number: str,
    firmware_version: str,
    timestamps: int,
    app_mask: bool = True,
    absolute_timestamps: bool = False,
    save_to_file: bool = False,
    correct_pix_address: bool = False,
    calculate_rates: bool = False,
) -> np.ndarray:
    """Collect data from files and apply mask to the valid pixel count.

    Unpacks data and returns the number of timestamps in each pixel.
    This function introduces modularity to the whole module and is
    called multiple times here.

    Parameters
    ----------
    files : List[str]
        List of data file paths.
    daughterboard_number : str
        The LinoSPAD2 daughterboard number.
    motherboard_number : str
        The LinoSPAD2 motherboard number, including the "#".
    firmware_version : str
        LinoSPAD2 firmware version.
    timestamps : int
        Number of timestamps per pixel per acquisition cycle.
    app_mask : bool, optional
        Switch for applying the mask on warm/hot pixels. Default is True.
    absolute_timestamps : bool, optional
        Indicator for data files with absolute timestamps. Default is
        False.
    correct_pix_address : bool, optional
        Check for correcting the pixel addresing. THe default is False.
    calculate_rates : bool, optional
        Switch for calculating the photon rate for each pixel. The
        default is 'False'.
    Returns
    -------
    np.ndarray
        Array with the number of timestamps per pixel.
    """
    # Define matrix of pixel coordinates, where rows are numbers of TDCs
    # and columns are the pixels that connected to these TDCs
    if firmware_version == "2212s":
        pix_coor = np.arange(256).reshape(4, 64).T
    elif firmware_version == "2212b":
        pix_coor = np.arange(256).reshape(64, 4)
    else:
        print("\nFirmware version is not recognized.")
        sys.exit()

    timestamps_per_pixel = np.zeros(256)

    # In the case a single file is passed, make a list out of it
    if isinstance(files, str):
        files = [files]

    for i in tqdm(range(len(files)), desc="Collecting data"):
        if not absolute_timestamps:
            data = f_up.unpack_binary_data(
                files[i],
                daughterboard_number,
                motherboard_number,
                firmware_version,
                timestamps,
                include_offset=False,
                apply_calibration=False,
            )
        else:
            data, _ = f_up.unpack_binary_data_with_absolute_timestamps(
                files[i],
                daughterboard_number,
                motherboard_number,
                firmware_version,
                timestamps,
                include_offset=False,
                apply_calibration=False,
            )
        for i in range(256):
            tdc, pix = np.argwhere(pix_coor == i)[0]
            ind = np.where(data[tdc].T[0] == pix)[0]
            ind1 = np.where(data[tdc].T[1][ind] > 0)[0]
            timestamps_per_pixel[i] += len(data[tdc].T[1][ind[ind1]])

    if correct_pix_address:
        fix = np.zeros(len(timestamps_per_pixel))
        fix[:128] = timestamps_per_pixel[128:]
        fix[128:] = np.flip(timestamps_per_pixel[:128])
        timestamps_per_pixel = fix
        del fix

    # Apply mask if requested
    if app_mask:
        mask = utils.apply_mask(daughterboard_number, motherboard_number)
        timestamps_per_pixel[mask] = 0

    acq_window_length = np.max(data[:].T[1]) * 1e-12
    number_of_cycles = len(np.where(data[0].T[0] == -2)[0])

    rates = (
        timestamps_per_pixel
        / acq_window_length
        / number_of_cycles
        / len(files)
    )

    if save_to_file:
        files.sort(key=os.path.getmtime)
        file_name = files[0][:-4] + "-" + files[-1][:-4]
        try:
            os.chdir("senpop_data")
        except FileNotFoundError as _:
            os.mkdir("senpop_data")
            os.chdir("senpop_data")

        np.savetxt(f"{file_name}_senpop_numbers.txt", timestamps_per_pixel)
        os.chdir("..")

    if calculate_rates:
        return timestamps_per_pixel, rates
    else:
        return timestamps_per_pixel


def plot_single_pix_hist(
    path,
    pixels,
    daughterboard_number: str,
    motherboard_number: str,
    firmware_version: str,
    timestamps: int = 512,
    cycle_length: float = 4e9,
    step: int = 1e6,
    show_fig: bool = False,
    fit_average: bool = False,
    color: str = "teal",
):
    """Plot a histogram for each pixel in the given range.

    Used mainly for checking the homogeneity of the LinoSPAD2 output
    (mainly clock and acquisition window size settings).

    Parameters
    ----------
    path : str
        Path to data file.
    pixels : array-like, list
        Array of pixels indices.
    daughterboard_number : str
        LinoSPAD2 daughterboard number.
    motherboard_number : str
        LinoSPAD2 motherboard (FPGA) number, including the '#'.
    firmware_version : str
        LinoSPAD2 firmware version.
    timestamps : int, optional
        Number of timestamps per pixel per acquisition cycle. The
        default is 512.
    cycle_length : float, optional
        Length of the data acquisition cycle. The default is 4e9, or 4 ms.
    step : int, optional
        Multiplier of 17.857 for the bin size of the histogram. The
        default is 1e6.
    show_fig : bool, optional
        Switch for showing the output figure. The default is False.
    fit_average : int, optional
        Switch for fitting averages of histogram counts in windows of
        +/-10. The default is False.
    color : str, optional
        Color of the histogram. The default is 'teal'.

    Returns
    -------
    None.

    """
    # parameter type check
    if isinstance(firmware_version, str) is not True:
        raise TypeError("'firmware_version' should be a string")
    if isinstance(daughterboard_number, str) is not True:
        raise TypeError("'daughterboard_number' should be a string")
    if isinstance(motherboard_number, str) is not True:
        raise TypeError("'motherboard_number' should be a string")

    def lin_fit(x, a, b):
        return a * x + b

    if type(pixels) is int:
        pixels = [pixels]

    os.chdir(path)

    # data_files = glob.glob("*.dat*")
    data_files = glob.glob("*.dat*")
    data_files.sort(key=os.path.getmtime)

    if show_fig is True:
        plt.ion()
    else:
        plt.ioff()
    for i, num in enumerate(data_files):
        print(f"> > > Plotting pixel histograms, Working on {num} < < <\n")

        data = f_up.unpack_binary_data(
            num,
            daughterboard_number,
            motherboard_number,
            firmware_version,
            timestamps,
            include_offset=False,
            apply_calibration=False,
        )

        bins = np.arange(
            0, cycle_length, 2500 / 140 * step
        )  # bin size of 17.867 us

        if pixels is None:
            pixels = np.arange(145, 165, 1)

        for i, _ in enumerate(pixels):
            plt.figure(figsize=(16, 10))
            plt.rcParams.update({"font.size": 27})
            # Define matrix of pixel coordinates, where rows are numbers
            # of TDCs and columns are the pixels that connected to
            # these TDCs
            if firmware_version == "2212s":
                pix_coor = np.arange(256).reshape(4, 64).T
            elif firmware_version == "2212b":
                pix_coor = np.arange(256).reshape(64, 4)
            else:
                print("\nFirmware version is not recognized.")
                sys.exit()
            tdc, pix = np.argwhere(pix_coor == pixels[i])[0]
            ind = np.where(data[tdc].T[0] == pix)[0]
            ind1 = np.where(data[tdc].T[1][ind] > 0)[0]
            data_to_plot = data[tdc].T[1][ind[ind1]]

            n, b, p = plt.hist(data_to_plot, bins=bins, color=color)
            if fit_average is True:
                av_win = np.zeros(int(len(n) / 10) + 1)
                av_win_in = np.zeros(int(len(n) / 10) + 1)
                for j, _ in enumerate(av_win):
                    av_win[j] = n[j * 10 : j * 10 + 1]
                    av_win_in[j] = b[j * 10 : j * 10 + 1]

                a = 1
                b = np.average(n)

                par, _ = curve_fit(lin_fit, av_win_in, av_win, p0=[a, b])

                av_win_fit = lin_fit(av_win_in, par[0], par[1])

            plt.xlabel("Time (ps)")
            plt.ylabel("Counts (-)")
            # plt.plot(av_win_in, av_win, color="black", linewidth=8)
            if fit_average is True:
                plt.gcf()
                plt.plot(av_win_in, av_win_fit, color="black", linewidth=8)
            plt.ticklabel_format(axis="y", style="sci", scilimits=(0, 0))
            plt.title(f"Pixel {pixels[i]}")
            try:
                os.chdir("results/single pixel histograms")
            except FileNotFoundError as _:
                os.makedirs("results/single pixel histograms")
                os.chdir("results/single pixel histograms")
            plt.savefig(f"{num}, pixel {pixels[i]}.png")
            os.chdir("../..")


def plot_sensor_population(
    path: str,
    daughterboard_number: str,
    motherboard_number: str,
    firmware_version: str,
    timestamps: int = 512,
    scale: str = "linear",
    style: str = "-o",
    show_fig: bool = False,
    app_mask: bool = True,
    color: str = "rebeccapurple",
    correct_pix_address: bool = False,
    fit_peaks: bool = False,
    threshold_multiplier: int = 10,
    pickle_fig: bool = False,
    absolute_timestamps: bool = False,
    single_file: bool = False,
) -> None:
    """Plot number of timestamps in each pixel for all datafiles.

    Plot sensor population as number of timestamps vs. pixel number.
    Analyzes all data files in the given folder. The output figure is saved
    in the "results" folder, which is created if it does not exist, in
    the same folder where datafiles are. Works with the firmware version
    '2212'.

    Parameters
    ----------
    path : str
        Path to the datafiles.
    daughterboard_number : str
        The LinoSPAD2 daughterboard number. Required for choosing the
        correct calibration data.
    motherboard_number : str
        The LinoSPAD2 motherboard number, including the "#".
    firmware_version : str
        LinoSPAD2 firmware version. Versions '2212b' (block) or '2212s'
        (skip) are recognized.
    timestamps : int, optional
        Number of timestamps per pixel per acquisition cycle. Default is
        "512".
    scale : str, optional
        Scale for the y-axis of the plot. Use "log" for logarithmic.
        The default is "linear".
    style : str, optional
        Style of the plot. The default is "-o".
    show_fig : bool, optional
        Switch for showing the plot. The default is False.
    app_mask : bool, optional
        Switch for applying the mask on warm/hot pixels. The default is
        True.
    color : str, optional
        Color for the plot. The default is 'rebeccapurple'.
    correct_pix_address : bool, optional
        Switch for correcting pixel addressing for the faulty firmware
        version for the 23 side of the daughterboard. The default is
        False.
    fit_peaks : bool, optional
        Switch for finding the highest peaks and fitting them with a
        Gaussian to provide their position. The default is False.
    threshold_multiplier : int, optional
        Threshold multiplier that is applied to median across the whole
        sensor for finding peaks. The default is 10.
    pickle_fig : bool, optional
        Switch for pickling the figure. Can be used when plotting takes
        a lot of time. The default is False.
    absolute_timestamps : bool, optional
        Indicator for data files with absolute timestamps. Default is
        False.
    single_file : optional
        Switch for using only the first file in the folder. Can be
        utilized for a quick plot. The default is False.

    Returns
    -------
    None.

    Examples
    -------
    An example how the function can be used to get the sensor
    occupation from a single file while looking for peaks - the most
    quick and straightforward approach to find where the beams were
    focused and get the peak position for further use in, e.g., delta_t
    functions. Here, the data were collected using the '23'-side
    sensor half which required correction of the pixel addressing.
    Offset calibration for the sensor is not available therefore
    it should be skipped.

    First, get the absolute path to where the '.dat' files are.
    >>> path = r'C:/Path/To/Data'

    Now to the function itself.
    >>> plot_sensor_popuation(
    >>> path,
    >>> daughterboard_number="NL11",
    >>> motherboard_number="#21",
    >>> firmware_version="2212s",
    >>> timestamps = 1000,
    >>> show_fig = True,
    >>> correct_pix_address = True,
    >>> fit_peaks = True,
    >>> single_file = True,
    >>> )
    """
    # parameter type check
    if not isinstance(firmware_version, str):
        raise TypeError(
            "'firmware_version' should be a string, '2212b' or '2212s'"
        )
    if not isinstance(daughterboard_number, str):
        raise TypeError(
            "'daughterboard_number' should be a string, 'NL11' or 'A5'"
        )
    if not isinstance(motherboard_number, str):
        raise TypeError("'motherboard_number' should be a string")
    if show_fig:
        plt.ion()
    else:
        plt.ioff()

    os.chdir(path)

    # files = glob.glob("*.dat*")
    files = glob.glob("*.dat*")
    files.sort(key=os.path.getmtime)

    if single_file:
        files = files[0]
        plot_name = files[:-4]
    else:
        plot_name = files[0][:-4] + "-" + files[-1][:-4]

    # valid_per_pixel = np.zeros(256)

    print(
        "\n> > > Collecting data for sensor population plot,"
        f"Working in {path} < < <\n"
    )

    # If fitting the peaks, calculate the photon rates in each peak as
    # well
    if fit_peaks:
        timestamps_per_pixel, rates = collect_data_and_apply_mask(
            files,
            daughterboard_number,
            motherboard_number,
            firmware_version,
            timestamps,
            app_mask,
            absolute_timestamps,
            save_to_file=False,
            correct_pix_address=correct_pix_address,
            calculate_rates=True,
        )
    else:
        timestamps_per_pixel = collect_data_and_apply_mask(
            files,
            daughterboard_number,
            motherboard_number,
            firmware_version,
            timestamps,
            app_mask,
            absolute_timestamps,
            save_to_file=False,
            correct_pix_address=correct_pix_address,
        )

    # Plotting
    print("\n> > > Plotting < < <\n")
    plt.rcParams.update({"font.size": 27})
    fig = plt.figure(figsize=(16, 10))
    if scale == "log":
        plt.yscale("log")
    plt.plot(timestamps_per_pixel, style, color=color)
    plt.xlabel("Pixel number (-)")
    plt.ylabel("Photons (-)")

    # Find and fit peaks if fit_peaks is True
    if fit_peaks:
        threshold = np.median(timestamps_per_pixel) * threshold_multiplier
        fit_width = 20
        peaks, _ = find_peaks(timestamps_per_pixel, height=threshold)
        peaks = np.unique(peaks)

        print("Fitting the peaks with gaussian")
        for peak_index in peaks:
            x_fit = np.arange(
                peak_index - fit_width, peak_index + fit_width + 1
            )
            cut_above_256 = np.where(x_fit >= 256)[0]
            x_fit = np.delete(x_fit, cut_above_256)
            y_fit = timestamps_per_pixel[x_fit]
            try:
                params, _ = utils.fit_gaussian(x_fit, y_fit)
            except Exception as _:
                continue

            plt.plot(
                x_fit,
                utils.gaussian(x_fit, *params),
                "--",
                label=f"Peak at {peak_index}, "
                f"Rate: {rates[peak_index]/1000:.0f} kHz",
            )

        plt.legend()

    # Save the figure
    try:
        os.chdir("results/sensor_population")
    except FileNotFoundError:
        os.makedirs("results/sensor_population")
        os.chdir("results/sensor_population")
    # fig.tight_layout()
    if single_file:
        plt.savefig(f"{plot_name}_single_file.png")
        print(
            "> > > The plot is saved as '{plot_name}_single_file.png'"
            "in {os.getcwd()} < < <"
        )
        if pickle_fig:
            pickle.dump(fig, open(f"{plot_name}_single_file.pickle", "wb"))
    else:
        plt.savefig(f"{plot_name}.png")
        print(
            f"> > > The plot is saved as '{plot_name}.png' "
            f"in {os.getcwd()} < < <"
        )
        if pickle_fig:
            pickle.dump(fig, open(f"{plot_name}.pickle", "wb"))

    os.chdir("../..")


# TODO don't have data to test the function
# def plot_sensor_population_spdc(
#     path,
#     daughterboard_number: str,
#     motherboard_number: str,
#     firmware_version: str,
#     timestamps: int = 512,
#     show_fig: bool = False,
# ):
#     """Plot sensor population for SPDC data.

#     Plots SPDC data subtracting the background (data with the SPDC
#     output turned off). Due to the low sensitivity of the LinoSPAD2
#     sensor to light at 810 nm of Thorlabs SPDC output, subtracting
#     background is required to show any meaningful signal.

#     Parameters
#     ----------
#     path : str
#         Path to data files.
#     daughterboard_number : str
#         LinoSPAD2 daughterboard number. Either "A5" or "NL11" is
#         accepted.
#     motherboard_number : str
#         LinoSPAD2 motherboard number.
#     timestamps : int, optional
#         Number of timestamps per pixel per acquisition cycle. The
#         default is 512.
#     show_fig : bool, optional
#         Switch for showing the plot. The default is False.

#     Raises
#     ------
#     TypeError
#         Raised when 'daughterboard_number' is not a string.
#     ValueError
#         Raised when the number of data files of SPDC data is different
#         from the number of data files of background data.

#     Returns
#     -------
#     None.

#     """
#     # parameter type check
#     if isinstance(daughterboard_number, str) is not True:
#         raise TypeError(
#             "'daughterboard_number' should be string, either 'NL11' or 'A5'"
#         )
#     if isinstance(firmware_version, str) is not True:
#         raise TypeError("'firmware_version' should be string")
#     if isinstance(motherboard_number, str) is not True:
#         raise TypeError(
#             "'daughterboard_number' should be string, either 'NL11' or 'A5'"
#         )

#     if show_fig is True:
#         plt.ion()
#     else:
#         plt.ioff()

#     os.chdir(path)

#     # files = glob.glob("*.dat*")

#     files = glob.glob("*.dat*")
#     files.sort(key=os.path.getmtime)

#     # background data for subtracting
#     path_bckg = path + "/bckg"
#     os.chdir(path_bckg)
#     files_bckg = glob.glob("*.dat*")
#     os.chdir("..")

#     if len(files) != len(files_bckg):
#         raise ValueError(
#             "Number of files with background data is different from the"
#             "number of actual data, exiting."
#         )

#     plot_name = files[0][:-4] + "-" + files[-1][:-4]

#     valid_per_pixel = np.zeros(256)
#     valid_per_pixel_bckg = np.zeros(256)

#     pix_coor = np.arange(256).reshape(64, 4)

#     # Collect SPDC data
#     for i in tqdm(range(len(files)), desc="Going through datafiles"):
#         data_all = f_up.unpack_binary_data(
#             files[i],
#             daughterboard_number="A5",
#             firmware_version="2212b",
#             timestamps=timestamps,
#         )

#         for i in np.arange(0, 256):
#             tdc, pix = np.argwhere(pix_coor == i)[0]
#             ind = np.where(data_all[tdc].T[0] == pix)[0]
#             ind1 = ind[np.where(data_all[tdc].T[1][ind] > 0)[0]]
#             valid_per_pixel[i] += len(data_all[tdc].T[1][ind1])

#     # Collect background data for subtracting
#     os.chdir(path_bckg)

#     for i in tqdm(
#         range(len(files_bckg)), desc="Going through background datafiles"
#     ):
#         data_all_bckg = f_up.unpack_binary_data(
#             files_bckg[i],
#             daughterboard_number="A5",
#             motherboard_number="34",
#             firmware_version="2212b",
#             timestamps=timestamps,
#         )

#         # Fot plotting counts
#         for i in np.arange(0, 256):
#             tdc, pix = np.argwhere(pix_coor == i)[0]
#             ind = np.where(data_all_bckg[tdc].T[0] == pix)[0]
#             ind1 = ind[np.where(data_all_bckg[tdc].T[1][ind] > 0)[0]]
#             valid_per_pixel_bckg[i] += len(data_all_bckg[tdc].T[1][ind1])

#     os.chdir("..")

#     # Mask the hot/warm pixels
#     path_to_back = os.getcwd()
#     path_to_mask = os.path.realpath(__file__) + "/../.." + "/params/masks"
#     os.chdir(path_to_mask)
#     file_mask = glob.glob(
#         "*{}_{}*".format(daughterboard_number, motherboard_number)
#     )[0]
#     mask = np.genfromtxt(file_mask).astype(int)
#     os.chdir(path_to_back)

#     for i in mask:
#         valid_per_pixel[i] = 0
#         valid_per_pixel_bckg[i] = 0

#     plt.rcParams.update({"font.size": 22})
#     plt.figure(figsize=(16, 10))
#     plt.xlabel("Pixel [-]")
#     plt.ylabel("Counts [-]")
#     plt.title("SPDC data, background subtracted")
#     plt.plot(valid_per_pixel - valid_per_pixel_bckg, "o-", color="teal")
#     plt.tight_layout()

#     try:
#         os.chdir("results/sensor_population")
#     except Exception:
#         os.makedirs("results/sensor_population")
#         os.chdir("results/sensor_population")
#     plt.savefig("{}_SPDC_counts.png".format(plot_name))
#     plt.pause(0.1)
#     os.chdir("../..")


def plot_sensor_population_full_sensor(
    path,
    daughterboard_number: str,
    motherboard_number1: str,
    motherboard_number2: str,
    firmware_version: str,
    timestamps: int = 512,
    scale: str = "linear",
    style: str = "-o",
    show_fig: bool = False,
    app_mask: bool = True,
    color: str = "salmon",
    fit_peaks: bool = False,
    threshold_multiplier: int = 10,
    pickle_fig: bool = False,
    single_file: bool = False,
    absolute_timestamps: bool = False,
):
    """Plot the number of timestamps in each pixel for all datafiles.

    Plot sensor population as the number of timestamps vs. pixel number.
    Analyzes all data files in the given folder. The output figure is saved
    in the "results" folder, which is created if it does not exist, in
    the same folder where datafiles are. Works with the firmware version
    '2212'.

    Parameters
    ----------
    path : str
        Path to the datafiles.
    daughterboard_number : str
        The LinoSPAD2 daughterboard number. Required for choosing the
        correct calibration data.
    motherboard_number1 : str
        The LinoSPAD2 motherboard number for the first board.
    motherboard_number2 : str
        The LinoSPAD2 motherboard number for the second board.
    firmware_version : str
        LinoSPAD2 firmware version. Versions '2212b' (block) or '2212s'
        (skip) are recognized.
    timestamps : int, optional
        Number of timestamps per pixel per acquisition cycle. Default is
        "512".
    scale : str, optional
        Scale for the y-axis of the plot. Use "log" for logarithmic.
        The default is "linear".
    style : str, optional
        Style of the plot. The default is "-o".
    show_fig : bool, optional
        Switch for showing the plot. The default is False.
    app_mask : bool, optional
        Switch for applying the mask on warm/hot pixels. The default is
        True.
    color : str, optional
        Color for the plot. The default is 'salmon'.
    fit_peaks : bool, optional
        Switch for finding the highest peaks and fitting them with a
        Gaussian to provide their position. The default is False.
    threshold_multiplier : int, optional
        Threshold multiplier for setting the threshold for finding peaks.
        The default is 10.
    pickle_fig : bool, optional
        Switch for pickling the figure. Can be used when plotting takes
        a lot of time. The default is False.
    single_file : bool, optional
        Switch for unpacking only the first file for a quick plot.
        The default is False.
    absolute_timestamps : bool, optional
        Indicator for data files with absolute timestamps. Default is
        False.

    Returns
    -------
    None.

    Notes
    -----

    As the pixel addressing is incorrect for one of the sensor halves
    (depends on the daughterboard-motherboards combinatios; in NL11, for
    motherboard #21 pixel addressing should be applied), it is important
    to apply pixel addressing to the correct board/sensor half. For this
    function, the order in which the motherboards are input is key, as
    pixel addressing is corrected for the second board.

    TLDR: motherboard_number2 should be for the one where correction of
    the pixel addressing is required.

    """
    # parameter type check
    if not isinstance(firmware_version, str):
        raise TypeError(
            "'firmware_version' should be a string, '2212b' or '2212s'"
        )
    if not isinstance(daughterboard_number, str):
        raise TypeError(
            "'daughterboard_number' should be a string, 'NL11' or 'A5'"
        )
    if not isinstance(motherboard_number1, str):
        raise TypeError("'motherboard_number1' should be a string")
    if not isinstance(motherboard_number2, str):
        raise TypeError("'motherboard_number2' should be a string")
    if show_fig:
        plt.ion()
    else:
        plt.ioff()

    valid_per_pixel1 = np.zeros(256)
    valid_per_pixel2 = np.zeros(256)

    # Get the two folders with data from both FPGAs/sensor halves
    os.chdir(path)
    path1 = glob.glob(f"*{motherboard_number1}*")[0]
    path2 = glob.glob(f"*{motherboard_number2}*")[0]

    # First motherboard / half of the sensor
    os.chdir(path1)
    # files1 = sorted(glob.glob("*.dat*"))

    files1 = glob.glob("*.dat*")
    files1.sort(key=os.path.getmtime)

    if single_file:
        files1 = files1[0]
    plot_name1 = files1[0][:-4] + "-"

    print(
        "\n> > > Collecting data for sensor population plot,"
        f"Working in {path1} < < <\n"
    )

    valid_per_pixel1 = collect_data_and_apply_mask(
        files1,
        daughterboard_number,
        motherboard_number1,
        firmware_version,
        timestamps,
        app_mask,
        absolute_timestamps,
    )

    os.chdir("..")

    # Second motherboard / half of the sensor
    os.chdir(path2)
    # files2 = sorted(glob.glob("*.dat*"))
    files2 = glob.glob("*.dat*")
    files2.sort(key=os.path.getmtime)
    if single_file:
        files2 = files2[0]
    plot_name2 = files2[-1][:-4]

    print(
        "\n> > > Collecting data for sensor population plot,"
        f"Working in {path2} < < <\n"
    )
    valid_per_pixel2 = collect_data_and_apply_mask(
        files2,
        daughterboard_number,
        motherboard_number2,
        firmware_version,
        timestamps,
        app_mask,
        absolute_timestamps,
    )

    # Fix pixel addressing for the second board
    fix = np.zeros(len(valid_per_pixel2))
    fix[:128] = valid_per_pixel2[128:]
    fix[128:] = np.flip(valid_per_pixel2[:128])
    valid_per_pixel2 = fix
    del fix

    # Concatenate and plot
    valid_per_pixel = np.concatenate([valid_per_pixel1, valid_per_pixel2])
    plot_name = plot_name1 + plot_name2

    print("\n> > > Plotting < < <\n")

    plt.rcParams.update({"font.size": 27})
    fig = plt.figure(figsize=(16, 10))
    if scale == "log":
        plt.yscale("log")
    plt.plot(valid_per_pixel, style, color=color)
    plt.xlabel("Pixel number (-)")
    plt.ylabel("Photons (-)")

    # Find and fit peaks if fit_peaks is True
    if fit_peaks:
        threshold = np.median(valid_per_pixel) * threshold_multiplier
        fit_width = 10
        peaks, _ = find_peaks(valid_per_pixel, height=threshold)
        peaks = np.unique(peaks)

        for peak_index in tqdm(peaks, desc="Fitting Gaussians"):
            x_fit = np.arange(
                peak_index - fit_width, peak_index + fit_width + 1
            )
            y_fit = valid_per_pixel[x_fit]
            try:
                params, _ = utils.fit_gaussian(x_fit, y_fit)
            except Exception:
                continue

            # amplitude, position, width = params
            # position = np.clip(int(position), 0, 255)

            plt.plot(
                x_fit,
                utils.gaussian(x_fit, *params),
                "--",
                label=f"Peak at {peak_index}",
            )

        plt.legend()

    os.chdir("..")

    try:
        os.chdir("results/sensor_population")
    except FileNotFoundError:
        os.makedirs("results/sensor_population")
        os.chdir("results/sensor_population")
    fig.tight_layout()
    plt.savefig("{}.png".format(plot_name))
    print(
        f"> > > The plot is saved as '{plot_name}.png' "
        f"in {os.getcwd()} < < <"
    )
    if pickle_fig:
        pickle.dump(fig, open(f"{plot_name}.pickle", "wb"))


def unpickle_plot(plot_pickle_file: str) -> dict:
    """Unpickle a saved figure and return plot data.

    Load a pickled figure with the sensor population plot, extract and
    return the plot data.

    Parameters
    ----------
    plot_pickle_file : str
        The absolute path to the pickle file with the sensor population
        plot.

    Returns
    -------
    tuple
        A tuple containing:
        - fig (matplotlib.figure.Figure): The unpickled figure object.
        - plot_data (dict): A dictionary with keys "Histogram" and
        "Fit_{i}" (for i >= 1), where each value is a tuple of x and y
        data for the corresponding line plot.

    Raises
    ------
    FileNotFoundError
        If the specified pickle file does not exist, a FileNotFoundError
        is raised and an error message is printed.
    """
    try:
        with open(plot_pickle_file, "rb") as f:
            fig = pickle.load(f)
    except FileNotFoundError as e:
        print(f" {e}")

    # Pack the data into a dictionary, first is the histogram, others
    # are the fits
    plot_data = {}

    # Extract the data; go over lines, stack them into dictionary
    ax = fig.axes[0]
    for i, line in enumerate(ax.lines):
        x, y = line.get_data()
        if i == 0:
            plot_data["Plot"] = (x, y)
        else:
            plot_data[f"Fit_{i}"] = (x, y)

    # Extract the legend
    legend = ax.get_legend()

    if legend is not None:
        legend_labels = [text.get_text() for text in legend.get_texts()]

        legend_labels = "\n".join(legend_labels)

        return fig, plot_data, legend_labels
    else:
        return fig, plot_data



"""Module for checking quality of data from LinoSPAD2.

The functions implemented in this module can be used to check data
quality, i.e., to study more closely how sensor population changes
between cycles or how the total number of timestamps in the given pixel
changes over time (cycles).

This file can also be imported as a module and contains the following
functions:

    * sensor_population_by_cycle - unpack data from the chosen file,
    collect number of timestamps in each pixel and cycle and plot the
    sensor population for either chosen cycles or for the starting
    cycle and 3 cycles on each side. The starting cycle can be found
    based on the given threshold, e.g., 0 for any positive signal or
    15 for a signal stronger than 3.5 kHz per pixel. The plots
    are then saved.

    * pixel_population_by_cycle - unpack data for all files in the given
    folder, collect number of timestamps accross all cycles for the
    given pixel and plot them against the cycle number. For clearer
    picture how the signal oscilates, also plot a moving average in a
    window of 100 cycles. The plot is then saved.

    * get_time_ratios - unpack data for all files in given folder,
    calculate the data collection time as the number of cycles in one
    file * the length of one cycle,  calculate the total acquisition
    time as the difference between the modification times of two
    subsequent files, calculate the ratios of these times and calculates
    the average with error.

    * save_file_times - save the modification and creation times of data
    files into a feather file for future use.

    * load_data_from_feather - load the modification and creation time
    of data files from the feather file.
    
    * sigma_of_count_spread_to_average - get timestamp differences from
    a '.feather' file, plot a histogram of background where no useful
    signal is, plot a jointplot of background signal including spread
    of the counts, plot and fit the spread using Gaussian.
    
    * sigma_of_count_spread_to_average_from_ft_file - get timestamp
    differences from a '.feather' file, plot a histogram of background
    where no useful signal is, plot a jointplot of background signal
    including spread of the counts, plot and fit the spread using
    Gaussian. Takes the feather file name directly and no raw data is
    required.

"""

import glob
import os
import sys
from datetime import datetime

import numpy as np
import pandas as pd
import pyarrow.feather as ft
import seaborn as sns
from matplotlib import pyplot as plt
from tqdm import tqdm

from LinoSPAD2.functions import unpack as f_up
from LinoSPAD2.functions import utils


def sensor_population_by_cycle(
    path: str,
    daughterboard_number: str,
    motherboard_number: str,
    firmware_version: str,
    timestamps: int = 512,
    app_mask: bool = True,
    include_offset: bool = False,
    apply_calibration: bool = True,
    absolute_timestamps: bool = False,
    correct_pixel_addressing: bool = False,
    cycle_range: list = None,
    threshold: int = 0,
    chosen_file: int = 0,
):
    """
    Collect sensor population data by acquisition cycle.

    Parameters
    ----------
    path : str
        Path to the data files.
    daughterboard_number : str
        LinoSPAD2 daughterboard number.
    motherboard_number : str
        LinoSPAD2 motherboard (FPGA) number, including the "#".
    firmware_version : str
        LinoSPAD2 firmware version. Versions '2212s' (skip) and '2212b'
        (block) are recognized.
    timestamps : int, optional
        Number of timestamps per acquisition cycle per pixel. The default
        is 512.
    app_mask : bool, optional
        Switch for applying pixel mask. The default is True.
    include_offset : bool, optional
        Switch for applying offset calibration. The default is True.
    apply_calibration : bool, optional
        Switch for applying TDC and offset calibration. If set to 'True'
        while include_offset is set to 'False', only the TDC calibration is
        applied. The default is True.
    absolute_timestamps : bool, optional
        Indicator for data with absolute timestamps. The default is
        False.
    correct_pixel_addressing : bool, optional
        Switch for correcting pixel addressing. The default is False.
    cycle_range : list, optional
        List of specific acquisition cycles to plot. The default is None.
    threshold : int, optional
        Threshold for finding the starting cycle: 0 for the first
        cycle where any signal appears. This parameter should be changed
        only in the case no cycles to plot were given manually. The
        default is 0.
    chosen_file : int, optional
        Index of the chosen file. The default is 0.

    Returns
    -------
    Tuple[np.ndarray, list]
        Tuple containing the sensor population array and the list of
        absolute timestamps.
    """

    os.chdir(path)

    files_all = glob.glob("*.dat*")
    # files_all.sort(key=lambda x: os.path.getctime(x))
    files_all.sort(key=os.path.getmtime)

    if firmware_version == "2212s":
        pix_coor = np.arange(256).reshape(4, 64).T
    elif firmware_version == "2212b":
        pix_coor = np.arange(256).reshape(64, 4)
    else:
        raise ValueError("Firmware version is not recognized.")

    if app_mask is True:
        mask = utils.apply_mask(daughterboard_number, motherboard_number)

    file = files_all[chosen_file]

    if not absolute_timestamps:
        data = f_up.unpack_binary_data(
            file,
            daughterboard_number,
            motherboard_number,
            firmware_version,
            timestamps,
            include_offset,
            apply_calibration,
        )
        sensor_population = np.zeros((256, int(len(data[0]) / (timestamps))))
    else:
        data, _ = f_up.unpack_binary_data_with_absolute_timestamps(
            file,
            daughterboard_number,
            motherboard_number,
            firmware_version,
            timestamps,
            include_offset,
            apply_calibration,
        )
        sensor_population = np.zeros(
            (256, int(len(data[0]) / (timestamps + 1)))
        )

    cycle_ends = np.where(data[0].T[1] == -2)[0]
    cycle_ends = np.insert(cycle_ends, 0, 0)
    indices = np.where(data[0].T[1])[0]

    for j in range(len(cycle_ends) - 1):
        cycle_indices = indices[
            (indices >= cycle_ends[j]) & (indices < cycle_ends[j + 1])
        ]
        for k in range(256):
            if k in mask:
                continue
            tdc, pix = np.argwhere(pix_coor == k)[0]
            pix_index = np.where(data[tdc].T[0][cycle_indices] == pix)[0]
            pix_index_pos = np.where(
                data[tdc].T[1][cycle_indices][pix_index] > 0
            )[0]

            sensor_population[k][j] += len(
                data[tdc].T[1][cycle_indices][pix_index[pix_index_pos]]
            )

    if correct_pixel_addressing:
        fix = np.zeros((256, 2200))
        fix[:128, :] = sensor_population[128:, :]
        fix[128:, :] = np.flip(sensor_population[:128, :])
        sensor_population = fix
        del fix

    try:
        os.chdir(os.path.join(path, "results", "data_quality", "senpop_cycle"))
    except FileNotFoundError:
        os.makedirs(
            os.path.join(path, "results", "data_quality", "senpop_cycle")
        )
        os.chdir(os.path.join(path, "results", "data_quality", "senpop_cycle"))

    # If cycles are not manually given, find the starting cycle
    if cycle_range is None:
        # Find the pixel where a peak is
        pix_peak = np.argmax(sensor_population[:, -1])
        # Find first cycle where height of the peak is above the threshold
        # which can be set to 15 for signal above 3.5 kHz
        cycle_start = np.where(sensor_population[pix_peak, :] > threshold)[
            0
        ].min()
        cycle_range = [x for x in range(cycle_start - 3, cycle_start + 3)]

    plt.rcParams.update({"font.size": 27})
    for _, cycle in enumerate(cycle_range):
        fig = plt.figure(figsize=(16, 10))
        plt.plot(sensor_population[:, cycle], "-.", color="rebeccapurple")
        plt.xlabel("Pixel index (-)")
        plt.ylabel("Sensor population (-)")
        plt.title(f"Cycle {cycle}")
        plt.tight_layout()

        # Save the figure
        fig.savefig(f"{os.path.splitext(file)[0]}_cycle_{cycle}.png")

        # Close all the figures to free up resources
        plt.close("all")


def pixel_population_by_cycle(
    path,
    pix_to_plot,
    daughterboard_number: str,
    motherboard_number: str,
    firmware_version: str,
    timestamps: int = 512,
    include_offset: bool = False,
    apply_calibration: bool = True,
    absolute_timestamps: bool = False,
    color: str = "rebeccapurple",
):
    """
    Collect and plot pixel population data by acquisition cycle.



    Parameters
    ----------
    path : str
        Path to the data files.
    pix_to_plot : int
        Pixel number to plot.
    daughterboard_number : str
        LinoSPAD2 daughterboard number.
    motherboard_number : str
        LinoSPAD2 motherboard (FPGA) number, including the "#".
    firmware_version : str
        LinoSPAD2 firmware version. Versions "2212s" (skip) and "2212b"
        (block) are recognized.
    timestamps : int, optional
        Number of timestamps per acquisition cycle per pixel. The default
        is 512.
    include_offset : bool, optional
        Switch for applying offset calibration. The default is True.
    apply_calibration : bool, optional
        Switch for applying TDC and offset calibration. If set to 'True'
        while include_offset is set to 'False', only the TDC calibration is
        applied. The default is True.
    absolute_timestamps : bool, optional
        Indicator for data with absolute timestamps. The default is
        False.
    color : str, optional
        Color for the plot. The default is "rebeccapurple".

    Returns
    -------
    None
    """
    os.chdir(path)

    files_all = glob.glob("*.dat*")
    # files_all.sort(key=lambda x: os.path.getmtime(x))
    files_all.sort(key=os.path.getmtime)

    # Define matrix of pixel coordinates, where rows are numbers of TDCs
    # and columns are the pixels that connected to these TDCs
    if firmware_version == "2212s":
        pix_coor = np.arange(256).reshape(4, 64).T
    elif firmware_version == "2212b":
        pix_coor = np.arange(256).reshape(64, 4)
    else:
        print("\nFirmware version is not recognized.")
        sys.exit()

    pixel_pop = []

    for i in tqdm(range(len(files_all)), desc="Collecting data"):
        # First board, unpack data
        file = files_all[i]

        if not absolute_timestamps:
            data = f_up.unpack_binary_data(
                file,
                daughterboard_number,
                motherboard_number,
                firmware_version,
                timestamps,
                include_offset,
                apply_calibration,
            )
        else:
            data, _ = f_up.unpack_binary_data_with_absolute_timestamps(
                file,
                daughterboard_number,
                motherboard_number,
                firmware_version,
                timestamps,
                include_offset,
                apply_calibration,
            )

        tdc, pix_c = np.argwhere(pix_coor == pix_to_plot)[0]
        pix = np.where(data[tdc].T[0] == pix_c)[0]

        column_data = data[:].T[1]

        cycle_ends = np.insert(np.where(data[tdc].T[1] == -2)[0], 0, 0)

        for i in range(len(cycle_ends) - 1):
            cycle_indices = pix[
                (pix >= cycle_ends[i]) & (pix < cycle_ends[i + 1])
            ]
            pixel_pop.append(
                len(column_data[cycle_indices][column_data[cycle_indices] > 0])
            )

    # Moving average of number of timestamps per 100 cycles
    moving_average = np.convolve(
        np.array(pixel_pop), np.ones(100) / 100, mode="valid"
    )

    # Plot moving average
    offset = (
        100 - 1
    ) // 2  # Offset for aligning moving average with original data

    try:
        os.chdir(os.path.join(path, "results", "data_quality", "pixpop_cycle"))
    except FileNotFoundError:
        os.makedirs(
            os.path.join(path, "results", "data_quality", "pixpop_cycle")
        )
        os.chdir(os.path.join(path, "results", "data_quality", "pixpop_cycle"))

    plt.rcParams.update({"font.size": 27})
    plt.figure(figsize=(16, 10))
    plt.plot(pixel_pop, "-.", color=color, label="Pixel population")
    plt.plot(
        range(offset, len(moving_average) + offset),
        moving_average,
        color="darkorange",
        label="Average in 100 cycles",
    )
    plt.xlabel("Acquisition cycle (-)")
    plt.ylabel("# of timestamps (-)")
    plt.legend()
    plt.title(f"Pixel {pix_to_plot}")

    plot_name = (
        os.path.splitext(files_all[0])[0]
        + "-"
        + os.path.splitext(files_all[-1])[0]
    )
    plt.savefig(f"{plot_name}_{pix_to_plot}.png")


def get_time_ratios(
    path: str,
    daughterboard_number: str,
    motherboard_number: str,
    firmware_version: str,
    timestamps: int,
    include_offset: bool = False,
    apply_calibration: bool = True,
):
    """
    Calculate the ratio of time of actual data collection and the time it
    takes to create and save the files.

    Parameters
    ----------
    Path : str
        Path to the binary data file.
    daughterboard_number : str
        LinoSPAD2 daughterboard number.
    motherboard_number : str
        LinoSPAD2 motherboard (FPGA) number, including the "#".
    firmware_version : str
        LinoSPAD2 firmware version. Either '2212s' (skip) or '2212b' (block).
    timestamps : int, optional
        Number of timestamps per cycle per TDC per acquisition cycle.
    include_offset : bool, optional
        Switch for applying offset calibration. The default is True.
    apply_calibration : bool, optional
        Switch for applying TDC and offset calibration. If set to 'True'
        while include_offset is set to 'False', only the TDC
        calibration is applied. The default is True.

    """

    os.chdir(path)

    files = glob.glob("*.dat*")
    files.sort(key=os.path.getmtime)

    timestamps_only = []
    number_of_cycles = []
    collecting_times = []

    tmsp_maxes = []

    for i in range(len(files) - 1):
        unpacked_data = f_up.unpack_binary_data(
            files[i],
            daughterboard_number,
            motherboard_number,
            firmware_version,
            timestamps,
            include_offset,
            apply_calibration,
        )

        timestamps_only = unpacked_data.T[1]

        tmsp_maxes.append(np.max(timestamps_only))

        number_of_cycles.append(len(np.where(unpacked_data[0].T[0] == -2)[0]))

        collecting_times.append(
            os.path.getmtime(files[i + 1]) - os.path.getmtime(files[i])
        )

    tp_max = max(tmsp_maxes) / 1e12

    ratios = [
        number_of_cycles[i] * tp_max / collecting_times[i]
        for i in range(len(collecting_times))
    ]

    ratio_average = np.mean(ratios)
    error = np.std(ratios) / np.sqrt(len(ratios))

    intercycle_time = [
        collecting_times[i] - number_of_cycles[i] * tp_max
        for i in range(len(collecting_times))
    ]

    print("Number of cycles:", number_of_cycles[1])
    print("Ratio average: ", ratio_average)
    print("Ratio error: ", error)


def save_file_times(path):
    """
    Save creation and modification times of data files into a .feather file.

    Parameters:

    path : str
        Path to the data files.

    """
    file_list = glob.glob(os.path.join(path, "*.dat*"))
    files_data = []

    file_list.sort()

    first_file_name = os.path.basename(file_list[0])[:-4]
    last_file_name = os.path.basename(file_list[-1])[:-4]

    for file_path in file_list:
        file_name = os.path.basename(file_path)
        creation_time = os.path.getctime(file_path)
        modification_time = os.path.getmtime(file_path)
        files_data.append(
            (
                file_name,
                datetime.fromtimestamp(creation_time),
                datetime.fromtimestamp(modification_time),
            )
        )

    df = pd.DataFrame(
        files_data, columns=["File Name", "Creation Time", "Modification Time"]
    )

    feather_filename = f"{first_file_name}_-_{last_file_name}.feather"
    ft.write_feather(df, os.path.join(path, feather_filename))

    return df


def load_data_from_feather(path):
    """
    Loads the creation and modification time from feather file.

    Parameters:

    path : str
        Path to the data files.

    """
    feather_files = glob.glob(os.path.join(path, "*.feather"))
    if feather_files:
        df_loaded = ft.read_feather(feather_files[0])
        return df_loaded
    else:
        print("No feather files found.")
        return None


def _extend_spread_range(spread_bins, spread_counts, extension: int):
    """Extend the background spread bins and counts.

    Can be used to extend the background histogram to both sides while
    keeping the same bin size and assigning zero counts to the newly
    added bins. Should improve the fit and, therefore, the sigma value.

    Parameters
    ----------
    spread_bins : array-like
        Original bins of the background spread histogram.
    spread_counts : array-like
        Original counts of the background spread histogram.
    extension : int
        Number of elements to add to the original bins and counts from
        each side.

    Returns
    -------
    array-like, array-like
        Extended bins and counts of the background spread histogram.
    """
    # Calculate the step size between elements
    step_size = spread_bins[1] - spread_bins[0]

    # Create arrays to add on both sides
    left_extension = np.arange(
        spread_bins[0] - extension * step_size,
        spread_bins[0],
        step_size,
    )
    right_extension = np.arange(
        spread_bins[-1] + step_size,
        spread_bins[-1] + (extension + 1) * step_size,
        step_size,
    )

    # Concatenate arrays
    extended_array = np.concatenate(
        (left_extension, spread_bins, right_extension)
    )
    extended_counts = np.concatenate(
        (np.zeros(extension), spread_counts, np.zeros(extension))
    )

    return extended_array, extended_counts


def sigma_of_count_spread_to_average(
    path: str,
    pixels: list,
    range_left: float = 20e3,
    range_right: float = 40e3,
    step: int = 10,
    bins_sigma: int = 20,
    extend: int = 0,
):
    """Plot and fit background spread from the feather file.

    Collect timestamp differences for the requested pixels from the
    given feather file, plot histogram of the background signal,
    plot a jointplot of the background with the spread of the
    background, fit the spread and calculate the ratio of the sigma of
    that spread to average background signal.

    Parameters
    ----------
    path : str
        Path to the data files.
    ft_file : str
        Feather file with timestamp differences.
    pixels : list
        Pixels which should be used for analysis.
    range_left : int, optional
        Lower limit for timestamp differences, lower values are not used.
        The default is 20e3.
    range_right : int, optional
        Upper limit for timestamp differences, higher values are not used.
        The default is 40e3.
    step : int, optional
        Multiplier of the average LinoSPAD2 TDC width of 17.857 ps that
        is used for histograms. The default is 10.
    bins_sigma : int, optional
        Number of bins used for plotting the histogram of the background
        spread. The default 20.
    extend: int, optional
        Number of elements to add to the background spread histogram
        for a better fit. The default is 0, when no elements are added.
    """
    os.chdir(path)

    ft_file = glob.glob("*.feather")[0]

    ft_file_name = ft_file.split(".")[0]

    data = ft.read_feather(ft_file)

    data_cut = data[f"{pixels[0]},{pixels[1]}"]

    # Cut the data from the background only; without the offset
    # calibration, the delta t peak rarely goes outside the 10 ns mark
    data_cut = data_cut[(data_cut > range_left) & (data_cut < range_right)]

    # Bins in units of 17.857 ps of the average LinoSPAD2 TDC bin width
    bins = np.arange(
        np.min(data_cut), np.max(data_cut), 2.5 / 140 * 1e3 * step
    )

    counts, bin_edges = np.histogram(data_cut, bins=bins)

    bin_centers = (bin_edges - 2.5 / 140 * 1e3 * step / 2)[1:]

    plt.rcParams.update({"font.size": 27})

    try:
        os.chdir("results/bckg_spread")
    except Exception:
        os.makedirs("results/bckg_spread")
        os.chdir("results/bckg_spread")

    # Background histogram
    plt.figure(figsize=(16, 10))
    plt.step(bin_centers, counts, color="tomato")
    plt.title(f"Histogram of delta ts\nBin size is {bins[1] - bins[0]:.2f} ps")
    plt.xlabel(r"$\Delta$t (ps)")
    plt.ylabel("# of coincidences (-)")
    plt.savefig(f"{ft_file_name}_bckg_hist.png")

    # Seaborn join histograms of background including the spread
    sns.jointplot(
        x=bin_centers, y=counts, height=10, marginal_kws=dict(bins=bins_sigma)
    )
    plt.title("Histogram of delta ts with histograms of spread", fontsize=27)
    plt.xlabel(r"$\Delta$t (ps)", fontsize=27)
    plt.ylabel("# of coincidences (-)", fontsize=27)
    plt.savefig(f"{ft_file_name}_bckg_hist_joint.png")

    # Histogram of the spread plus Gaussian fit
    counts_spread, bin_edges_spread = np.histogram(counts, bins=bins_sigma)
    bin_centers_spread = (
        bin_edges_spread - (bin_edges_spread[1] - bin_edges_spread[0]) / 2
    )[1:]

    if extend > 0:
        bin_centers_spread, counts_spread = _extend_spread_range(
            bin_centers_spread, counts_spread, extend
        )

    pars, covs = utils.fit_gaussian(bin_centers_spread, counts_spread)

    fig, ax = plt.subplots(figsize=(16, 10))
    ax.step(
        bin_centers_spread,
        counts_spread,
        label="Spread of counts",
        color="darkorchid",
    )
    ax.plot(
        bin_centers_spread,
        utils.gaussian(bin_centers_spread, *pars),
        label="Fit",
        color="#cc8c32",
    )
    ax.set_title(
        f"Ratio of spread to average: {pars[2] / np.mean(counts) * 100:.1f} %"
    )
    ax.set_xlabel("Spread (-)")
    ax.set_ylabel("Counts (-)")
    ax.text(
        0.07,
        0.9,
        f"\u03C3={pars[2]:.2f}\u00B1{np.sqrt(covs[2,2]):.2f}",
        transform=ax.transAxes,
        fontsize=25,
        bbox=dict(
            facecolor="white", edgecolor="black", boxstyle="round,pad=0.5"
        ),
    )
    plt.savefig(f"{ft_file_name}_bckg_spread_hist_.png")


def sigma_of_count_spread_to_average_from_ft_file(
    path: str,
    ft_file: str,
    pixel_pair: list,
    range_left: float = 20e3,
    range_right: float = 40e3,
    step: int = 10,
    bins_sigma: int = 20,
    extend: int = 0,
):
    """Plot and fit background spread from the feather file.

    Collect timestamp differences for the requested pixels from the
    given feather file, plot histogram of the background signal,
    plot a jointplot of the background with the spread of the
    background, fit the spread and calculate the ratio of the sigma of
    that spread to average background signal. Only the feather file
    is required without the raw data and the feather file is chosen
    directly.

    Parameters
    ----------
    path : str
        Path to where the feather file is.
    ft_file : str
        Feather file with timestamp differences.
    pixel_pair : list
        Two pixels which should be used for analysis.
    range_left : int, optional
        Lower limit for timestamp differences, lower values are not used.
        The default is 20e3.
    range_right : int, optional
        Upper limit for timestamp differences, higher values are not used.
        The default is 40e3.
    step : int, optional
        Multiplier of the average LinoSPAD2 TDC width of 17.857 ps that
        is used for histograms. The default is 10.
    bins_sigma : int, optional
        Number of bins used for plotting the histogram of the background
        spread. The default 20.
    extend: int, optional
        Number of elements to add to the background spread histogram
        for a better fit. The default is 0, when no elements are added.
    """

    ft_file_name = ft_file.split(".")[0]

    os.chdir(path)

    data = ft.read_feather(ft_file)

    data_cut = data[f"{pixel_pair[0]},{pixel_pair[1]}"]

    # Cut the data from the background only; without the offset
    # calibration, the delta t peak rarely goes outside the 10 ns mark
    data_cut = data_cut[(data_cut > range_left) & (data_cut < range_right)]

    # Bins in units of 17.857 ps of the average LinoSPAD2 TDC bin width
    bins = np.arange(
        np.min(data_cut), np.max(data_cut), 2.5 / 140 * 1e3 * step
    )

    counts, bin_edges = np.histogram(data_cut, bins=bins)

    bin_centers = (bin_edges - 2.5 / 140 * 1e3 * step / 2)[1:]

    plt.rcParams.update({"font.size": 27})

    try:
        os.chdir("results/bckg_spread")
    except FileNotFoundError:
        os.makedirs("results/bckg_spread")
        os.chdir("results/bckg_spread")

    # Background histogram
    plt.figure(figsize=(16, 10))
    plt.step(bin_centers, counts, color="tomato")
    plt.title(f"Histogram of delta ts\nBin size is {bins[1] - bins[0]:.2f} ps")
    plt.xlabel(r"$\Delta$t (ps)")
    plt.ylabel("# of coincidences (-)")
    plt.savefig(f"{ft_file_name}_bckg_hist.png")

    # Seaborn join histograms of background including the spread
    sns.jointplot(
        x=bin_centers, y=counts, height=10, marginal_kws=dict(bins=bins_sigma)
    )
    plt.title("Histogram of delta ts with histograms of spread", fontsize=27)
    plt.xlabel(r"$\Delta$t (ps)", fontsize=27)
    plt.ylabel("# of coincidences (-)", fontsize=27)
    plt.savefig(f"{ft_file_name}_bckg_hist_joint.png")

    # Histogram of the spread plus Gaussian fit
    counts_spread, bin_edges_spread = np.histogram(counts, bins=bins_sigma)
    bin_centers_spread = (
        bin_edges_spread - (bin_edges_spread[1] - bin_edges_spread[0]) / 2
    )[1:]

    # Extend the range (if required) for the background spread for a
    # better fit
    if extend > 0:
        bin_centers_spread, counts_spread = _extend_spread_range(
            bin_centers_spread, counts_spread, extend
        )

    pars, covs = utils.fit_gaussian(bin_centers_spread, counts_spread)

    fig, ax = plt.subplots(figsize=(16, 10))
    ax.step(
        bin_centers_spread,
        counts_spread,
        label="Spread of counts",
        color="darkorchid",
    )
    ax.plot(
        bin_centers_spread,
        utils.gaussian(bin_centers_spread, *pars),
        label="Fit",
        color="#cc8c32",
    )
    ax.set_title(
        f"Ratio of spread to average: {pars[2] / np.mean(counts) * 100:.1f} %"
    )
    ax.set_xlabel("Spread (-)")
    ax.set_ylabel("Counts (-)")
    ax.text(
        0.59,
        0.9,
        f"\u03C3={pars[2]:.2f}\u00B1{np.sqrt(covs[2,2]):.2f}",
        transform=ax.transAxes,
        fontsize=25,
        bbox=dict(
            facecolor="white", edgecolor="black", boxstyle="round,pad=0.5"
        ),
    )
    plt.savefig(f"{ft_file_name}_bckg_spread_hist_.png")



import os
import shutil
import unittest

import numpy as np

from LinoSPAD2.functions.delta_t import (
    calculate_and_save_timestamp_differences,
    calculate_and_save_timestamp_differences_fast,
    collect_and_plot_timestamp_differences,
)
from LinoSPAD2.functions.fits import (
    fit_with_gaussian,
    fit_with_gaussian_all,
    fit_with_gaussian_fancy,
)


class TestDeltasFull(unittest.TestCase):
    def setUp(self):
        # Set up test variables
        self.partial_path = "tests/test_data"
        self.pixels = [
            [x for x in range(66, 70)],
            [x for x in range(170, 178)],
        ]
        self.daughterboard_number = "NL11"
        self.motherboard_number = "#33"
        self.firmware_version = "2212b"
        self.timestamps = 300
        self.delta_window = 20e3
        self.rewrite = True
        self.range_left = -20e3
        self.range_right = 20e3
        self.same_y = False
        self.app_mask = True
        self.include_offset = False

    def test_a_deltas_save_positive(self):
        # Test positive case for deltas_save function
        work_dir = os.path.dirname(os.path.realpath(__file__)) + "/.."

        # os.chdir(work_dir)

        path = os.path.join(work_dir, self.partial_path)
        calculate_and_save_timestamp_differences(
            path,
            self.pixels,
            self.rewrite,
            self.daughterboard_number,
            self.motherboard_number,
            self.firmware_version,
            self.timestamps,
            self.delta_window,
            self.app_mask,
            self.include_offset,
        )

        os.chdir(
            os.path.join(
                os.path.dirname(os.path.realpath(__file__)), "test_data"
            )
        )

        # Check if the csv file is created
        self.assertTrue(
            os.path.isfile(
                "delta_ts_data/test_data_2212b-test_data_2212b.feather"
            )
        )

    def test_a_deltas_save_fast_positive(self):
        # Test positive case for deltas_save function
        work_dir = os.path.dirname(os.path.realpath(__file__)) + "/.."

        # os.chdir(work_dir)

        path = os.path.join(work_dir, self.partial_path)
        calculate_and_save_timestamp_differences_fast(
            path,
            self.pixels,
            self.rewrite,
            self.daughterboard_number,
            self.motherboard_number,
            self.firmware_version,
            self.timestamps,
            self.delta_window,
            app_mask=self.app_mask,
            include_offset=self.include_offset,
        )

        os.chdir(
            os.path.join(
                os.path.dirname(os.path.realpath(__file__)), "test_data"
            )
        )

        # Check if the csv file is created
        self.assertTrue(
            os.path.isfile(
                "delta_ts_data/test_data_2212b-test_data_2212b.feather"
            )
        )

    # Negative test case
    # Invalid firmware version
    def test_b_deltas_save_negative(self):
        work_dir = os.path.dirname(os.path.realpath(__file__)) + "/.."

        # os.chdir(work_dir)
        path = os.path.join(work_dir, self.partial_path)

        # Test negative case for deltas_save function
        with self.assertRaises(TypeError):
            calculate_and_save_timestamp_differences(
                path,
                self.pixels,
                "2212",
                self.daughterboard_number,
                self.motherboard_number,
                self.firmware_version,
                self.timestamps,
                self.delta_window,
            )

    def test_c_delta_cp(self):
        # Test case for delta_cp function
        # Positive test case
        work_dir = os.path.dirname(os.path.realpath(__file__)) + "/.."

        # os.chdir(work_dir)
        path = os.path.join(work_dir, self.partial_path)

        collect_and_plot_timestamp_differences(
            path,
            pixels=[x for x in range(67, 69)] + [x for x in range(173, 175)],
            rewrite=self.rewrite,
            range_left=self.range_left,
            range_right=self.range_right,
            same_y=self.same_y,
        )

        # Check if the plot file is created
        self.assertTrue(
            os.path.isfile(
                "results/delta_t/test_data_2212b-test_data_2212b_delta_t_grid.png"
            )
        )

    def test_d_fit_with_gaussian_positive(self):
        # Test with valid input
        work_dir = os.path.dirname(os.path.realpath(__file__)) + "/.."

        path = os.path.join(work_dir, self.partial_path)
        pixels = [67, 174]
        window = 20e3
        multiplier = 5

        # Call the function
        fit_with_gaussian(
            path,
            pixels,
            ft_file=None,
            window=window,
            multiplier=multiplier,
        )

        # Assert that the function runs without raising any exceptions
        self.assertTrue(
            os.path.isfile(
                "results/fits/test_data_2212b-test_data_2212b_"
                f"pixels_{pixels[0]},{pixels[1]}_fit.png"
            )
        )

    def test_d_fit_with_gaussian_all_positive(self):
        # Test with valid input
        pixels = [82, 116]

        work_dir = os.path.dirname(os.path.realpath(__file__)) + "/.."

        path = os.path.join(work_dir, self.partial_path)

        ft_file = r"test.feather"

        # Call the function
        fit_with_gaussian_all(
            path,
            pixels=pixels,
            ft_file=ft_file,
            window=10e3,
            multiplier=5,
        )

        # Assert that the function runs without raising any exceptions
        self.assertTrue(
            os.path.isfile(
                f"results/fits/test_pixels_{pixels[0]},{pixels[1]}_all_fit.png"
            )
        )

    def test_d_fit_with_gaussian_fancy_positive(self):
        # Test with valid input
        pixels = [82, 116]

        work_dir = os.path.dirname(os.path.realpath(__file__)) + "/.."

        path = os.path.join(work_dir, self.partial_path)

        ft_file = r"test.feather"

        # Call the function
        fit_with_gaussian_fancy(
            path,
            pixels=pixels,
            ft_file=ft_file,
            range_left=-5e3,
            range_right=5e3,
            multiplier=5,
        )

        # Assert that the function runs without raising any exceptions
        self.assertTrue(
            os.path.isfile(
                f"results/fits/test_pixels_{pixels[0]},{pixels[1]}_fancy_fit.png"
            )
        )

    def test_d_fit_with_gaussian_pickle_positive(self):
        # Test with valid input
        work_dir = os.path.dirname(os.path.realpath(__file__)) + "/.."

        path = os.path.join(work_dir, self.partial_path)
        pixels = [67, 174]
        window = 20e3
        multiplier = 5

        # Call the function
        fit_with_gaussian(
            path,
            pixels,
            ft_file=None,
            window=window,
            multiplier=multiplier,
            pickle_figure=True,
        )

        # Assert that the function runs without raising any exceptions
        self.assertTrue(
            os.path.isfile(
                "results/fits/test_data_2212b-test_data_2212b_"
                f"pixels_{pixels[0]},{pixels[1]}_fit.pkl"
            )
        )

    def test_d_fit_with_gaussian_all_pickle_positive(self):
        # Test with valid input
        pixels = [82, 116]

        work_dir = os.path.dirname(os.path.realpath(__file__)) + "/.."

        path = os.path.join(work_dir, self.partial_path)

        ft_file = r"test.feather"

        # Call the function
        fit_with_gaussian_all(
            path,
            pixels=pixels,
            ft_file=ft_file,
            window=10e3,
            multiplier=5,
            pickle_figure=True,
        )

        # Assert that the function runs without raising any exceptions
        self.assertTrue(
            os.path.isfile(
                f"results/fits/test_pixels_{pixels[0]},{pixels[1]}_all_fit.pkl"
            )
        )

    def test_d_fit_with_gaussian_fancy_pickle_positive(self):
        # Test with valid input
        pixels = [82, 116]

        work_dir = os.path.dirname(os.path.realpath(__file__)) + "/.."

        path = os.path.join(work_dir, self.partial_path)

        ft_file = r"test.feather"

        # Call the function
        fit_with_gaussian_fancy(
            path,
            pixels=pixels,
            ft_file=ft_file,
            range_left=-5e3,
            range_right=5e3,
            multiplier=5,
            pickle_figure=True,
        )

        # Assert that the function runs without raising any exceptions
        self.assertTrue(
            os.path.isfile(
                f"results/fits/test_pixels_{pixels[0]},{pixels[1]}_fancy_fit.pkl"
            )
        )

    def tearDownClass():
        # Clean up after tests
        os.chdir(r"{}".format(os.path.dirname(os.path.realpath(__file__))))
        shutil.rmtree("test_data/delta_ts_data")
        shutil.rmtree("test_data/results")


if __name__ == "__main__":
    unittest.main()



import glob
import os
import shutil
import unittest

import numpy as np
import pandas as pd

from LinoSPAD2.functions.cross_talk import (
    calculate_dark_count_rate,
    collect_cross_talk,
    plot_cross_talk,
)


class TestCTFull(unittest.TestCase):
    def setUp(self):
        # Set up test variables
        self.path = "tests/test_data"
        self.pixels = [x for x in range(0, 20)]
        self.daughterboard_number = "NL11"
        self.motherboard_number = "#33"
        self.firmware_version = "2212s"
        self.timestamps = 300
        self.delta_window = 10e3
        self.step = 1
        self.pix1 = 0
        self.scale = "linear"
        self.include_offset = False

    def test_a_collect_ct_positive(self):
        work_dir = r"{}".format(
            os.path.dirname(os.path.realpath(__file__)) + "/.."
        )
        os.chdir(work_dir)
        # Test positive case of collect_ct function
        collect_cross_talk(
            self.path,
            self.pixels,
            self.daughterboard_number,
            self.motherboard_number,
            self.firmware_version,
            self.timestamps,
            self.delta_window,
            self.step,
            self.include_offset,
        )
        # Check if the output file is created and has the correct number of rows
        file = glob.glob("*CT_data_*.csv*")[0]
        data = pd.read_csv(file, header=None)
        self.assertEqual(len(data), 20)

    def test_b_collect_ct_negative(self):
        work_dir = r"{}".format(
            os.path.dirname(os.path.realpath(__file__)) + "/.."
        )
        os.chdir(work_dir)
        # Test negative case of collect_ct function
        with self.assertRaises(TypeError):
            collect_cross_talk(
                self.path,
                self.pixels,
                123,
                22,
                self.timestamps,
                self.delta_window,
            )

    def test_c_plot_ct_positive(self):
        # Test positive case of plot_ct function
        work_dir = r"{}".format(
            os.path.dirname(os.path.realpath(__file__)) + "/.."
        )
        os.chdir(work_dir)
        plot_cross_talk(self.path, self.pix1, self.scale)
        # Check if the plot file is created
        plot_name = "test_data_2212b.dat_test_data_2212b.dat"
        plot_file = "{plot}_{pix}.png".format(plot=plot_name, pix=self.pix1)
        self.assertTrue(os.path.exists(plot_file))

    def test_d_plot_ct_negative(self):
        # Test negative case of plot_ct function
        with self.assertRaises(FileNotFoundError):
            plot_cross_talk("nonexistent_folder", self.pix1, self.scale)

    def test_calculate_dark_count_rate_positive(self):
        work_dir = r"{}".format(
            os.path.dirname(os.path.realpath(__file__)) + "/.."
        )
        os.chdir(work_dir)
        result = calculate_dark_count_rate(
            self.path,
            self.daughterboard_number,
            self.motherboard_number,
            self.firmware_version,
            self.timestamps,
        )
        # Add your assertions here based on expected results
        assert isinstance(result, float)
        assert result >= 0  # Assuming the result should be non-negative

    def tearDownClass():
        # Clean up after tests
        os.chdir(r"{}".format(os.path.dirname(os.path.realpath(__file__))))
        shutil.rmtree("test_data/cross_talk_data")
        shutil.rmtree("test_data/results")


if __name__ == "__main__":
    unittest.main()
